from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Optional

import hypothesis.strategies as st
import narwhals as nw
from annotated_types import Interval
from hypothesis import assume, given
from pydantic import BaseModel, FiniteFloat, NegativeFloat, NonNegativeFloat, NonPositiveFloat, PositiveFloat

from tests.pydantic.utils import model_to_nw_schema

if TYPE_CHECKING:
    from anyschema.parsers import ParserPipeline


@given(lb=st.floats(), ub=st.floats())
def test_parse_float(auto_pipeline: ParserPipeline, lb: float, ub: float) -> None:
    assume(lb < ub)

    class FloatModel(BaseModel):
        # python float type
        py_int: float
        py_float_optional: float | None
        py_float_or_none: float | None
        none_or_py_float: None | float

        # pydantic NonNegativeFloat type
        non_negative: NonNegativeFloat
        non_negative_optional: NonNegativeFloat | None
        non_negative_or_none: NonNegativeFloat | None
        none_or_non_negative: None | NonNegativeFloat

        # pydantic NonPositiveFloat type
        non_positive: NonPositiveFloat
        non_positive_optional: NonPositiveFloat | None
        non_positive_or_none: NonPositiveFloat | None
        none_or_non_positive: None | NonPositiveFloat

        # pydantic PositiveFloat type
        positive: PositiveFloat
        positive_optional: PositiveFloat | None
        positive_or_none: PositiveFloat | None
        none_or_positive: None | PositiveFloat

        # pydantic NegativeFloat type
        negative: NegativeFloat
        negative_optional: NegativeFloat | None
        negative_or_none: NegativeFloat | None
        none_or_negative: None | NegativeFloat

        # pydantic NegativeFloat type
        finite: FiniteFloat
        finite_optional: FiniteFloat | None
        finite_or_none: FiniteFloat | None
        none_or_finite: None | NegativeFloat

        # pydantic annotated float with constraints
        con_float: Annotated[float, Interval(gt=lb, lt=ub)]
        con_float_optional: Optional[Annotated[float, Interval(ge=lb, lt=ub)]]
        con_float_or_none: Annotated[float, Interval(gt=lb, le=ub)] | None
        non_or_con_float: None | Annotated[float, Interval(ge=lb, le=ub)]

    schema = model_to_nw_schema(FloatModel, pipeline=auto_pipeline)

    assert all(value == nw.Float64() for value in schema.values())



from __future__ import annotations

from anyschema.parsers._annotated import AnnotatedStep
from anyschema.parsers._base import ParserStep
from anyschema.parsers._builtin import PyTypeStep
from anyschema.parsers._forward_ref import ForwardRefStep
from anyschema.parsers._pipeline import ParserPipeline, make_pipeline
from anyschema.parsers._union import UnionTypeStep

__all__ = (
    "AnnotatedStep",
    "ForwardRefStep",
    "ParserPipeline",
    "ParserStep",
    "PyTypeStep",
    "UnionTypeStep",
    "make_pipeline",
)



from __future__ import annotations

import sys
from collections.abc import Mapping, Sequence
from dataclasses import is_dataclass as dc_is_dataclass
from importlib.util import find_spec
from typing import TYPE_CHECKING

from typing_extensions import TypeIs, is_typeddict

if TYPE_CHECKING:
    from types import ModuleType

    from pydantic import BaseModel

    from anyschema.typing import AttrsClassType, DataclassType, IntoOrderedDict, SQLAlchemyTableType, TypedDictType

ANNOTATED_TYPES_AVAILABLE = find_spec("annotated_types") is not None
PYDANTIC_AVAILABLE = find_spec("pydantic") is not None
ATTRS_AVAILABLE = find_spec("attrs") is not None
SQLALCHEMY_AVAILABLE = find_spec("sqlalchemy") is not None


def get_pydantic() -> ModuleType | None:
    """Get pydantic module (if already imported - else return None)."""
    return sys.modules.get("pydantic", None)


def get_attrs() -> ModuleType | None:
    """Get attrs module (if already imported - else return None)."""
    return sys.modules.get("attrs", None)


def is_into_ordered_dict(obj: object) -> TypeIs[IntoOrderedDict]:
    """Check if the object can be converted into a python OrderedDict."""
    tpl_size = 2
    return isinstance(obj, Mapping) or (
        isinstance(obj, Sequence) and all(isinstance(s, tuple) and len(s) == tpl_size for s in obj)
    )


def is_typed_dict(obj: object) -> TypeIs[TypedDictType]:
    """Check if the object is a TypedDict and narrows type checkers."""
    return is_typeddict(obj)


def is_dataclass(obj: object) -> TypeIs[DataclassType]:
    """Check if the object is a dataclass and narrows type checkers."""
    return dc_is_dataclass(obj)


def is_pydantic_base_model(obj: object) -> TypeIs[type[BaseModel]]:
    """Check if the object is a pydantic BaseModel."""
    return (
        (pydantic := get_pydantic()) is not None
        and isinstance(obj, type)
        and isinstance(obj, type(pydantic.BaseModel))
        and issubclass(obj, pydantic.BaseModel)
    )


def is_attrs_class(obj: object) -> TypeIs[AttrsClassType]:
    """Check if the object is an attrs class.

    Uses attrs.has() to check if a class is an attrs class.
    Supports @attrs.define/@attrs.frozen decorators.
    """
    return (attrs := get_attrs()) is not None and attrs.has(obj)


def get_sqlalchemy() -> ModuleType | None:
    """Get sqlalchemy module (if already imported - else return None)."""
    return sys.modules.get("sqlalchemy", None)


def get_sqlalchemy_orm() -> ModuleType | None:
    """Get sqlalchemy.orm module (if already imported - else return None)."""
    return sys.modules.get("sqlalchemy.orm", None)


def is_sqlalchemy_table(obj: object) -> TypeIs[SQLAlchemyTableType]:
    """Check if the object is a SQLAlchemy Table or DeclarativeBase class.

    Supports both:

    - SQLAlchemy Table instances (Core)
    - SQLAlchemy ORM mapped classes (DeclarativeBase subclasses)
    """
    is_table = (sql := get_sqlalchemy()) is not None and isinstance(obj, sql.Table)
    is_declarative_base = (
        (sql_orm := get_sqlalchemy_orm()) is not None
        and isinstance(obj, type)
        and issubclass(obj, sql_orm.DeclarativeBase)
    )
    return is_table or is_declarative_base



from __future__ import annotations

from typing import TYPE_CHECKING

import narwhals as nw

from anyschema._dependencies import is_attrs_class
from anyschema.parsers._base import ParserStep

if TYPE_CHECKING:
    from narwhals.dtypes import DType

    from anyschema.typing import AttrsClassType, FieldConstraints, FieldMetadata, FieldType


__all__ = ("AttrsTypeStep",)


class AttrsTypeStep(ParserStep):
    """Parser for attrs-specific types.

    Handles:

    - attrs classes (Struct types)

    Warning:
        It requires [attrs](https://www.attrs.org/) to be installed.
    """

    def parse(
        self,
        input_type: FieldType,
        constraints: FieldConstraints,  # noqa: ARG002
        metadata: FieldMetadata,  # noqa: ARG002
    ) -> DType | None:
        """Parse attrs-specific types into Narwhals dtypes.

        Arguments:
            input_type: The type to parse.
            constraints: Constraints associated with the type.
            metadata: Custom metadata dictionary.

        Returns:
            A Narwhals DType if this parser can handle the type, None otherwise.
        """
        if is_attrs_class(input_type):
            return self._parse_attrs_class(input_type)

        # This parser doesn't handle this type
        return None

    def _parse_attrs_class(self, attrs_class: AttrsClassType) -> DType:
        """Parse an attrs class into a Struct type.

        Arguments:
            attrs_class: The attrs class.

        Returns:
            A Narwhals Struct dtype.
        """
        from anyschema.adapters import attrs_adapter

        return nw.Struct(
            [
                nw.Field(
                    name=field_name,
                    dtype=self.pipeline.parse(field_type, field_constraints, field_metadata, strict=True),
                )
                for field_name, field_type, field_constraints, field_metadata in attrs_adapter(attrs_class)
            ]
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

import pytest
from pydantic import AwareDatetime, create_model

from anyschema.exceptions import UnsupportedDTypeError
from tests.pydantic.utils import model_to_nw_schema

if TYPE_CHECKING:
    from anyschema.parsers import ParserPipeline


@pytest.mark.parametrize(
    ("input_type", "msg"),
    [
        (str | float | int, "Union with more than two types is not supported."),
        (str | float, "Union with mixed types is not supported."),
    ],
)
def test_raise_parse_union(auto_pipeline: ParserPipeline, input_type: type, msg: str) -> None:
    ExceptionModel = create_model("ExceptionModel", foo=(input_type, ...))  # noqa: N806

    with pytest.raises(UnsupportedDTypeError, match=msg):
        model_to_nw_schema(ExceptionModel, pipeline=auto_pipeline)


@pytest.mark.parametrize(
    "input_type",
    [
        AwareDatetime,
        Optional[AwareDatetime],
        AwareDatetime | None,
        None | AwareDatetime,
    ],
)
def test_raise_aware_datetime(auto_pipeline: ParserPipeline, input_type: type) -> None:
    AwareDatetimeModel = create_model("AwareDatetimeModel", foo=(input_type, ...))  # noqa: N806

    msg = "pydantic AwareDatetime does not specify a fixed timezone."
    with pytest.raises(UnsupportedDTypeError, match=msg):
        model_to_nw_schema(AwareDatetimeModel, pipeline=auto_pipeline)



from __future__ import annotations

from typing import Any, TypeAlias

import pytest

from anyschema import AnySchema

AnyDict: TypeAlias = dict[str, Any]


@pytest.mark.parametrize(
    ("spec", "expected"),
    [
        ({}, ()),
        ({"id": int, "name": str}, (False, False)),
        ({"id": int, "name": str, "age": int}, (False, False, False)),
        ({"id": int, "name": str, "age": int | None}, (False, False, True)),
    ],
)
def test_nullables_named_false(spec: AnyDict, expected: tuple[bool, ...]) -> None:
    schema = AnySchema(spec=spec)
    result = schema.nullables()

    assert result == expected


@pytest.mark.parametrize(
    ("spec", "expected"),
    [
        ({}, {}),
        ({"id": int, "name": str}, (False, False)),
        ({"id": int, "name": str, "age": int}, (False, False, False)),
        ({"id": int, "name": str, "age": int | None}, (False, False, True)),
    ],
)
def test_nullables_named_true(spec: AnyDict, expected: tuple[bool, ...]) -> None:
    schema = AnySchema(spec=spec)
    result = schema.nullables(named=True)

    assert result == dict(zip(spec, expected, strict=True))



from __future__ import annotations

import enum

import pytest
from narwhals.dtypes import (
    Array,
    Binary,
    Boolean,
    Categorical,
    Date,
    Datetime,
    Decimal,
    DType,
    Duration,
    Enum,
    Float32,
    Float64,
    Int8,
    Int16,
    Int32,
    Int64,
    Int128,
    List,
    Object,
    String,
    Struct,
    Time,
    UInt8,
    UInt16,
    UInt32,
    UInt64,
    UInt128,
    Unknown,
)

from anyschema.exceptions import UnsupportedDTypeError
from anyschema.serde import _extract_field_name, deserialize_dtype, serialize_dtype


class Color(enum.Enum):
    BLUE = 1
    RED = 2


@pytest.mark.parametrize(
    ("dtype", "expected"),
    [
        (Array(Int32(), 2), "Array(Int32, shape=(2,))"),
        (Array(Int32(), (2, 3)), "Array(Int32, shape=(2, 3))"),
        (Array(List(Int32()), (2, 3)), "Array(List(Int32), shape=(2, 3))"),
        (Binary(), "Binary"),
        (Boolean(), "Boolean"),
        (Categorical(), "Categorical"),
        (Date(), "Date"),
        (Datetime(), "Datetime(time_unit='us', time_zone=None)"),
        (Datetime(time_unit="ms"), "Datetime(time_unit='ms', time_zone=None)"),
        (Datetime(time_zone="UTC"), "Datetime(time_unit='us', time_zone='UTC')"),
        (Datetime("s", "Europe/Berlin"), "Datetime(time_unit='s', time_zone='Europe/Berlin')"),
        (Decimal(), "Decimal"),
        (Duration(), "Duration(time_unit='us')"),
        (Duration(time_unit="ms"), "Duration(time_unit='ms')"),
        (Enum(Color), "Enum(categories=[1, 2])"),
        (Enum(["blue", "red"]), "Enum(categories=['blue', 'red'])"),
        (Float32(), "Float32"),
        (Float64(), "Float64"),
        (Int8(), "Int8"),
        (Int16(), "Int16"),
        (Int32(), "Int32"),
        (Int64(), "Int64"),
        (Int128(), "Int128"),
        (List(Int32()), "List(Int32)"),
        (List(Struct({"int": Int32(), "str": String()})), "List(Struct({'int': Int32, 'str': String}))"),
        (Object(), "Object"),
        (String(), "String"),
        (Struct({}), "Struct({})"),
        (Struct({"int": Int32(), "str": String()}), "Struct({'int': Int32, 'str': String})"),
        (
            Struct({"lst": List(Int32()), "arr": Array(Int16(), (3, 2))}),
            "Struct({'lst': List(Int32), 'arr': Array(Int16, shape=(3, 2))})",
        ),
        (
            Struct({"struct": Struct({"enum": Enum(["blue", "red"])})}),
            "Struct({'struct': Struct({'enum': Enum(categories=['blue', 'red'])})})",
        ),
        (Time(), "Time"),
        (UInt8(), "UInt8"),
        (UInt16(), "UInt16"),
        (UInt32(), "UInt32"),
        (UInt64(), "UInt64"),
        (UInt128(), "UInt128"),
        (Unknown(), "Unknown"),
    ],
)
def test_serialize(dtype: DType, expected: str) -> None:
    result = serialize_dtype(dtype)
    assert result == expected

    # round-trip
    assert deserialize_dtype(result) == dtype


@pytest.mark.parametrize(
    ("into_dtype", "expected"),
    [
        ("Array(Int32, shape=(2,))", Array(Int32(), 2)),
        ("Array(Int32, shape=(2, 3))", Array(Int32(), (2, 3))),
        ("Array(List(Int32), shape=(2, 3))", Array(List(Int32()), (2, 3))),
        ("Binary", Binary()),
        ("Boolean", Boolean()),
        ("Categorical", Categorical()),
        ("Date", Date()),
        ("Datetime(time_unit='us', time_zone=None)", Datetime()),
        ("Datetime(time_unit='ms', time_zone=None)", Datetime(time_unit="ms")),
        ("Datetime(time_unit='us', time_zone='UTC')", Datetime(time_zone="UTC")),
        ("Datetime(time_unit='s', time_zone='Europe/Berlin')", Datetime("s", "Europe/Berlin")),
        ("Decimal", Decimal()),
        ("Duration(time_unit='us')", Duration()),
        ("Duration(time_unit='ms')", Duration(time_unit="ms")),
        ("Enum(categories=[1, 2])", Enum(Color)),
        ("Enum(categories=['blue', 'red'])", Enum(["blue", "red"])),
        ("Float32", Float32()),
        ("Float64", Float64()),
        ("Int8", Int8()),
        ("Int16", Int16()),
        ("Int32", Int32()),
        ("Int64", Int64()),
        ("Int128", Int128()),
        ("List(Int32)", List(Int32())),
        ("List(Struct({'int': Int32, 'str': String}))", List(Struct({"int": Int32(), "str": String()}))),
        ("Object", Object()),
        ("String", String()),
        ("Struct({})", Struct({})),
        ("Struct({'int': Int32, 'str': String})", Struct({"int": Int32(), "str": String()})),
        (
            "Struct({'lst': List(Int32), 'arr': Array(Int16, shape=(3, 2))})",
            Struct({"lst": List(Int32()), "arr": Array(Int16(), (3, 2))}),
        ),
        (
            "Struct({'struct': Struct({'enum': Enum(categories=['blue', 'red'])})})",
            Struct({"struct": Struct({"enum": Enum(["blue", "red"])})}),
        ),
        ("Time", Time()),
        ("UInt8", UInt8()),
        ("UInt16", UInt16()),
        ("UInt32", UInt32()),
        ("UInt64", UInt64()),
        ("UInt128", UInt128()),
        ("Unknown", Unknown()),
    ],
)
def test_deserialize(into_dtype: str, expected: DType) -> None:
    result = deserialize_dtype(into_dtype)
    assert result == expected

    # round-trip
    assert serialize_dtype(result) == into_dtype


@pytest.mark.parametrize("into_dtype", ["InvalidType", "List(", "not a dtype at all"])
def test_deserialize_invalid_dtype_raises(into_dtype: str) -> None:
    with pytest.raises(UnsupportedDTypeError, match="Unable to deserialize"):
        deserialize_dtype(into_dtype)


@pytest.mark.parametrize("field_str", ["fieldname: Int64", "123: Int64"])
def test_extract_field_name_invalid_raises(field_str: str) -> None:
    with pytest.raises(ValueError, match="Failed to parse field name"):
        _extract_field_name(field_str, start_pos=0)



from __future__ import annotations

from typing import TYPE_CHECKING

import narwhals as nw
from pydantic import AwareDatetime, BaseModel, FutureDate, FutureDatetime, NaiveDatetime, PastDate, PastDatetime

from anyschema._dependencies import is_pydantic_base_model
from anyschema._metadata import get_anyschema_value_by_key
from anyschema.exceptions import UnsupportedDTypeError
from anyschema.parsers._base import ParserStep

if TYPE_CHECKING:
    from narwhals.dtypes import DType

    from anyschema.typing import FieldConstraints, FieldMetadata, FieldType

__all__ = ("PydanticTypeStep",)


class PydanticTypeStep(ParserStep):
    """Parser for Pydantic-specific types.

    Handles:

    - Pydantic datetime types (`AwareDatetime`, `NaiveDatetime`, etc.)
    - Pydantic date types (`PastDate`, `FutureDate`)
    - Pydantic `BaseModel` (Struct types)

    Warning:
        It requires [pydantic](https://docs.pydantic.dev/latest/) to be installed.
    """

    def parse(
        self,
        input_type: FieldType,
        constraints: FieldConstraints,  # noqa: ARG002
        metadata: FieldMetadata,
    ) -> DType | None:
        """Parse Pydantic-specific types into Narwhals dtypes.

        Arguments:
            input_type: The type to parse.
            constraints: Constraints associated with the type.
            metadata: Custom metadata dictionary.

        Returns:
            A Narwhals DType if this parser can handle the type, None otherwise.
        """
        # Check if it's a type/class first (not a generic alias or other special form)
        if not isinstance(input_type, type):
            return None

        # Handle AwareDatetime
        if issubclass(input_type, AwareDatetime):  # pyright: ignore[reportArgumentType]  # ty: ignore[invalid-argument-type]
            # Pydantic AwareDatetime does not fix a single timezone, but any timezone would work.
            # See https://github.com/pydantic/pydantic/issues/5829
            # Unless a timezone is specified via {"anyschema": {"time_zone": ...}}, we raise an error.
            if (time_zone := get_anyschema_value_by_key(metadata, key="time_zone")) is None:
                msg = (
                    "pydantic AwareDatetime does not specify a fixed timezone.\n\n"
                    "Hint: You can specify a timezone via "
                    "`Field(..., json_schema_extra={'anyschema': {'time_zone': 'UTC'}})`"
                )
                raise UnsupportedDTypeError(msg)

            return nw.Datetime(
                time_unit=get_anyschema_value_by_key(metadata, key="time_unit", default="us"), time_zone=time_zone
            )

        if issubclass(input_type, NaiveDatetime):  # pyright: ignore[reportArgumentType] # ty: ignore[invalid-argument-type]
            # Pydantic NaiveDatetime should not receive a timezone.
            # If a timezone is specified via {"anyschema": {"time_zone": ...}}, we raise an error.
            if (time_zone := get_anyschema_value_by_key(metadata, key="time_zone")) is not None:
                msg = f"pydantic NaiveDatetime should not specify a timezone, found {time_zone}."
                raise UnsupportedDTypeError(msg)

            return nw.Datetime(
                time_unit=get_anyschema_value_by_key(metadata, key="time_unit", default="us"), time_zone=None
            )

        # Handle datetime types
        if issubclass(input_type, (PastDatetime, FutureDatetime)):  # pyright: ignore[reportArgumentType]
            # PastDatetime and FutureDatetime accept both aware and naive datetimes.
            return nw.Datetime(
                time_unit=get_anyschema_value_by_key(metadata, key="time_unit", default="us"),
                time_zone=get_anyschema_value_by_key(metadata, key="time_zone"),
            )

        # Handle date types
        if issubclass(input_type, (PastDate, FutureDate)):  # pyright: ignore[reportArgumentType]
            return nw.Date()

        # Handle Pydantic models (Struct types)
        if is_pydantic_base_model(input_type):
            return self._parse_pydantic_model(input_type)

        # TODO(FBruzzesi): Add support for more pydantic types. See https://github.com/FBruzzesi/anyschema/issues/45

        # This parser doesn't handle this type
        return None

    def _parse_pydantic_model(self, model: type[BaseModel]) -> DType:
        """Parse a Pydantic model into a Struct type.

        Arguments:
            model: The Pydantic model class or instance.

        Returns:
            A Narwhals Struct dtype.
        """
        from anyschema.adapters import pydantic_adapter

        return nw.Struct(
            [
                nw.Field(
                    name=field_name,
                    dtype=self.pipeline.parse(field_info, field_constraints, field_metadata, strict=True),
                )
                for field_name, field_info, field_constraints, field_metadata in pydantic_adapter(model)
            ]
        )



from __future__ import annotations

import sys
from dataclasses import dataclass, field, make_dataclass
from datetime import date, datetime
from typing import TYPE_CHECKING

import pytest
from pydantic.dataclasses import dataclass as pydantic_dataclass

from anyschema.adapters import dataclass_adapter
from tests.conftest import DataclassEventWithTimeMetadata

if TYPE_CHECKING:
    from anyschema.typing import DataclassType, FieldSpec


class PersonIntoDataclass:
    name: str
    age: int
    date_of_birth: date


@pytest.mark.parametrize(
    "spec",
    [
        pydantic_dataclass(PersonIntoDataclass),
        dataclass(PersonIntoDataclass),
        make_dataclass("Test", [("name", str), ("age", int), ("date_of_birth", date)]),
    ],
)
def test_dataclass_adapter(spec: DataclassType) -> None:
    expected: tuple[FieldSpec, ...] = (("name", str, (), {}), ("age", int, (), {}), ("date_of_birth", date, (), {}))
    result = tuple(dataclass_adapter(spec))
    assert result == expected


def test_dataclass_adapter_missing_decorator_raises() -> None:
    """Test that adapter raises helpful error when child class isn't decorated."""

    @dataclass
    class Base:
        foo: str

    class ChildWithoutDecorator(Base):
        bar: int

    expected_msg = (
        "Class 'ChildWithoutDecorator' has annotations ('bar') that are not dataclass fields. "
        "If this class inherits from a dataclass, you must also decorate it with @dataclass "
        "to properly define these fields."
    )

    with pytest.raises(AssertionError, match=expected_msg.replace("(", r"\(").replace(")", r"\)")):
        list(dataclass_adapter(ChildWithoutDecorator))  # ty: ignore[invalid-argument-type]


def test_dataclass_adapter_with_time_metadata() -> None:
    result = tuple(dataclass_adapter(DataclassEventWithTimeMetadata))  # ty: ignore[invalid-argument-type]

    expected: tuple[FieldSpec, ...] = (
        ("name", str, (), {"anyschema": {"description": "Event name"}}),
        ("created_at", datetime, (), {}),
        ("scheduled_at", datetime, (), {"anyschema": {"time_zone": "UTC", "description": "Scheduled time"}}),
        ("started_at", datetime, (), {"anyschema": {"time_unit": "ms"}}),
        ("completed_at", datetime, (), {"anyschema": {"time_zone": "Europe/Berlin", "time_unit": "ns"}}),
    )

    assert result == expected


@pytest.mark.skipif(sys.version_info < (3, 14), reason="doc parameter requires Python 3.14+")
def test_dataclass_adapter_with_doc_argument() -> None:
    @dataclass
    class Product:
        name: str = field(doc="Product name")  # pyright: ignore[reportCallIssue]  # ty: ignore[no-matching-overload]
        price: float = field(  # pyright: ignore[reportCallIssue]  # ty: ignore[no-matching-overload]
            doc="Product price",
            metadata={"anyschema": {"description": "From metadata"}},  # anyschema metadata have precedence
        )
        in_stock: bool

    result = list(dataclass_adapter(Product))  # ty: ignore[invalid-argument-type]
    expected = [
        ("name", str, (), {"anyschema": {"description": "Product name"}}),
        ("price", float, (), {"anyschema": {"description": "From metadata"}}),
        ("in_stock", bool, (), {}),
    ]
    assert result == expected



# ruff: noqa: T201
from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, TypeVar

if TYPE_CHECKING:
    from typing_extensions import TypeIs

_T = TypeVar("_T")


def qualified_type_name(obj: object | type[Any], /) -> str:
    # Copied from Narwhals: https://github.com/narwhals-dev/narwhals/blob/282a3cb08f406e2f319d86b81a7300a2a6c5f390/narwhals/_utils.py#L1922
    # Author: Marco Gorelli
    # License: MIT: https://github.com/narwhals-dev/narwhals/blob/282a3cb08f406e2f319d86b81a7300a2a6c5f390/LICENSE.md
    tp = obj if isinstance(obj, type) else type(obj)
    module = tp.__module__ if tp.__module__ != "builtins" else ""
    return f"{module}.{tp.__name__}".lstrip(".")


def _get_sys_info() -> dict[str, str]:
    """System information.

    Returns system and Python version information

    Adapted from sklearn.

    Returns:
        Dictionary with system info.
    """
    import platform
    import sys

    python = sys.version.replace("\n", " ")

    blob = (
        ("python", python),
        ("machine", platform.platform()),
    )

    return dict(blob)


def _get_deps_info() -> dict[str, str]:
    """Overview of the installed version of main dependencies.

    This function does not import the modules to collect the version numbers
    but instead relies on standard Python package metadata.

    Returns version information on relevant Python libraries

    This function and show_versions were copied from sklearn and adapted

    Returns:
        Mapping from dependency to version.
    """
    from importlib.metadata import distributions

    libs = (
        "anyschema",
        "narwhals",
        "typing_extensions",
        "attrs",
        "pydantic",
        "sqlalchemy",
        "pandas",
        "polars",
        "pyarrow",
    )
    dist_map = {dist.name.lower(): dist.version for dist in distributions()}
    return {lib: dist_map.get(lib, "") for lib in libs}


def show_versions() -> None:
    """Print useful debugging information.

    Examples:
        >>> from anyschema import show_versions
        >>> show_versions()  # doctest: +SKIP
    """
    sys_info = _get_sys_info()
    deps_info = _get_deps_info()

    print("\nSystem:")
    for k, stat in sys_info.items():
        print(f"{k:>10}: {stat}")

    print("\nPython dependencies:")
    for k, stat in deps_info.items():
        print(f"{k:>20}: {stat}")


def is_sequence_but_not_str(sequence: Sequence[_T] | Any) -> TypeIs[Sequence[_T]]:
    return isinstance(sequence, Sequence) and not isinstance(sequence, str)


def is_sequence_of(obj: Any, tp: type[_T]) -> TypeIs[Sequence[_T]]:
    # Check if an object is a sequence of `tp`, only sniffing the first element.
    return bool(is_sequence_but_not_str(obj) and (first := next(iter(obj), None)) and isinstance(first, tp))



"""Tests for derived types similar to pydantic-extra-types.

This module tests that PyTypeStep correctly handles types that inherit
from basic Python types, similar to those in pydantic-extra-types library.

The derived types mimic pydantic-extra-types

References:
- https://docs.pydantic.dev/latest/api/pydantic_extra_types_country/
- https://docs.pydantic.dev/latest/api/pydantic_extra_types_phone_numbers/
- https://docs.pydantic.dev/latest/api/pydantic_extra_types_routing_numbers/
"""

from __future__ import annotations

from datetime import date, datetime
from decimal import Decimal
from enum import Enum
from typing import Any

import narwhals as nw
import pytest

from anyschema.parsers import ParserPipeline, PyTypeStep, UnionTypeStep


class EmailStr(str):
    __slots__ = ()


class PositiveInt(int): ...


class PositiveFloat(float): ...


class SecretBytes(bytes): ...


class PastDatetime(datetime): ...


class FutureDate(date): ...


class NonNegativeDecimal(Decimal): ...


class HttpMethod(str, Enum):
    """HTTP method enum."""

    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"


class Priority(int, Enum):
    """Priority levels."""

    LOW = 1
    MEDIUM = 2
    HIGH = 3


@pytest.fixture(scope="module")
def parser_pipeline() -> ParserPipeline:
    """Create a parser pipeline with UnionTypeStep and PyTypeStep."""
    union_parser = UnionTypeStep()
    py_parser = PyTypeStep()
    return ParserPipeline([union_parser, py_parser])


@pytest.mark.parametrize(
    ("input_type", "expected"),
    [
        (EmailStr, nw.String()),
        (EmailStr | None, nw.String()),
        (list[EmailStr], nw.List(nw.String())),
        (PositiveInt, nw.Int64()),
        (list[PositiveInt], nw.List(nw.Int64())),
        (PositiveFloat, nw.Float64()),
        (tuple[PositiveFloat, PositiveFloat], nw.Array(nw.Float64(), shape=2)),
        (SecretBytes, nw.Binary()),
        (PastDatetime, nw.Datetime("us")),
        (FutureDate, nw.Date),
        (NonNegativeDecimal, nw.Decimal()),
        (HttpMethod, nw.Enum(HttpMethod)),
        (Priority, nw.Enum(Priority)),
    ],
)
def test_derived_types(parser_pipeline: ParserPipeline, input_type: Any, expected: nw.dtypes.DType) -> None:
    result = parser_pipeline.parse(input_type, constraints=(), metadata={})
    assert result == expected



"""Tests for pydantic parser with derived types.

This module tests that PydanticTypeStep correctly handles types that inherit
from Pydantic's date/datetime types.
"""

from __future__ import annotations

from typing import Any

import narwhals as nw
import pytest
from pydantic import FutureDate, PastDate, PastDatetime

from anyschema.parsers import ParserPipeline, PyTypeStep, UnionTypeStep
from anyschema.parsers.pydantic import PydanticTypeStep


# Custom types that inherit from Pydantic types
class CustomPastDate(PastDate): ...


class CustomFutureDate(FutureDate): ...


class CustomPastDatetime(PastDatetime): ...


@pytest.fixture(scope="module")
def parser_pipeline() -> ParserPipeline:
    """Create a parser pipeline with pydantic support."""
    union_parser = UnionTypeStep()
    pydantic_parser = PydanticTypeStep()
    py_parser = PyTypeStep()
    return ParserPipeline([union_parser, pydantic_parser, py_parser])


@pytest.mark.parametrize(
    ("input_type", "expected"),
    [
        # Base Pydantic types
        (PastDate, nw.Date()),
        (FutureDate, nw.Date()),
        (PastDatetime, nw.Datetime()),
        # Derived types
        (CustomPastDate, nw.Date()),
        (CustomFutureDate, nw.Date()),
        (CustomPastDatetime, nw.Datetime()),
        # In lists
        (list[CustomPastDate], nw.List(nw.Date())),
        (list[CustomPastDatetime], nw.List(nw.Datetime())),
    ],
)
def test_pydantic_derived_types(parser_pipeline: ParserPipeline, input_type: Any, expected: nw.dtypes.DType) -> None:
    """Test that pydantic parser handles derived types correctly."""
    result = parser_pipeline.parse(input_type, (), {})
    assert result == expected



from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Optional

import narwhals as nw
import pydantic
import pytest
from narwhals.utils import parse_version
from pydantic import BaseModel, StrictStr

from tests.pydantic.utils import model_to_nw_schema

if TYPE_CHECKING:
    from anyschema.parsers import ParserPipeline


def test_parse_string(auto_pipeline: ParserPipeline) -> None:
    class StringModel(BaseModel):
        # python str type
        py_str: str
        py_str_optional: str | None
        py_str_or_none: str | None
        none_or_py_str: None | str

        # pydantic StrictStr type
        strict_str: StrictStr
        strict_str_optional: StrictStr | None
        strict_str_or_none: StrictStr | None
        none_or_strict_str: None | StrictStr

    schema = model_to_nw_schema(StringModel, pipeline=auto_pipeline)

    assert all(value == nw.String() for value in schema.values())


@pytest.mark.skipif(parse_version(pydantic.__version__) < (2, 1), reason="too old for StringConstraints")
def test_parse_string_with_constraints(auto_pipeline: ParserPipeline) -> None:
    from pydantic import StringConstraints

    str_constraint = StringConstraints(strip_whitespace=True, to_upper=True, pattern=r"^[A-Z]+$")

    class StringConstraintsModel(BaseModel):
        str_con: Annotated[str, str_constraint]
        str_con_optional: Optional[Annotated[str, str_constraint]]
        str_con_or_none: Annotated[str, str_constraint] | None
        none_or_str_con: None | Annotated[str, str_constraint]

    schema = model_to_nw_schema(StringConstraintsModel, pipeline=auto_pipeline)

    assert all(value == nw.String() for value in schema.values())



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import date, datetime
from typing import Literal, Sequence

import attrs
import narwhals as nw
import pytest
from pydantic import AwareDatetime, BaseModel, Field, FutureDatetime, NaiveDatetime, PastDate, PastDatetime, PositiveInt
from sqlalchemy import ARRAY, BigInteger, Boolean, Column, Date, DateTime, Float, Integer, MetaData, String, Table
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from anyschema.parsers import ParserPipeline


@pytest.fixture(scope="session")
def auto_pipeline() -> ParserPipeline:
    """Fixture to get the auto pipeline."""
    return ParserPipeline("auto")


@pytest.fixture
def nw_schema() -> nw.Schema:
    return nw.Schema(
        {
            "array": nw.Array(nw.Int32(), 3),
            "boolean": nw.Boolean(),
            "categorical": nw.Categorical(),
            "date": nw.Date(),
            "datetime": nw.Datetime(),
            "decimal": nw.Decimal(),
            "duration": nw.Duration(),
            "enum": nw.Enum(["foo", "bar"]),
            "float32": nw.Float32(),
            "float64": nw.Float64(),
            "int8": nw.Int8(),
            "int16": nw.Int16(),
            "int32": nw.Int32(),
            "int64": nw.Int64(),
            "int128": nw.Int128(),
            "list": nw.List(nw.Float32()),
            "object": nw.Object(),
            "string": nw.String(),
            "struct": nw.Struct(fields=[nw.Field("field_1", nw.String()), nw.Field("field_2", nw.Boolean())]),
            "uint8": nw.UInt8(),
            "uint16": nw.UInt16(),
            "uint32": nw.UInt32(),
            "uint64": nw.UInt64(),
            "uint128": nw.UInt128(),
            "unknown": nw.Unknown(),
        }
    )


class PydanticStudent(BaseModel):
    name: str = Field(description="Student full name")
    date_of_birth: PastDate
    age: PositiveInt = Field(description="Student age in years")
    classes: list[str]
    has_graduated: bool


@pytest.fixture(scope="session")
def pydantic_student_cls() -> type[PydanticStudent]:
    return PydanticStudent


class PydanticEventWithTimeMetadata(BaseModel):
    """Pydantic model with datetime fields that have time metadata."""

    name: str
    created_at: datetime
    scheduled_at: datetime = Field(json_schema_extra={"anyschema": {"time_zone": "UTC"}})
    started_at: datetime = Field(json_schema_extra={"anyschema": {"time_unit": "ms"}})
    completed_at: datetime = Field(json_schema_extra={"anyschema": {"time_zone": "Europe/Berlin", "time_unit": "ns"}})


class PydanticSpecialDatetimeWithMetadata(BaseModel):
    """Pydantic model with special datetime types and metadata."""

    aware: AwareDatetime = Field(json_schema_extra={"anyschema": {"time_zone": "UTC"}})
    aware_ms: AwareDatetime = Field(json_schema_extra={"anyschema": {"time_zone": "Asia/Tokyo", "time_unit": "ms"}})
    naive_ms: NaiveDatetime = Field(json_schema_extra={"anyschema": {"time_unit": "ms"}})
    past_utc: PastDatetime = Field(json_schema_extra={"anyschema": {"time_zone": "UTC"}})
    future_ns: FutureDatetime = Field(json_schema_extra={"anyschema": {"time_unit": "ns"}})


class PydanticEventWithXAnyschema(BaseModel):
    """Pydantic model using x-anyschema prefix (OpenAPI compatibility)."""

    name: str
    created_at: datetime = Field(json_schema_extra={"x-anyschema": {"time_zone": "UTC"}})
    started_at: datetime = Field(json_schema_extra={"x-anyschema": {"time_unit": "ms"}})


@attrs.define
class AttrsAddress:
    street: str
    city: str


@attrs.define
class AttrsPerson:
    name: str
    age: int
    date_of_birth: date
    is_active: bool
    classes: list[str]
    grades: list[float]


@attrs.frozen
class AttrsPersonFrozen:
    name: str
    age: int
    date_of_birth: date


@attrs.define
class AttrsBase:
    foo: str
    bar: int


@attrs.define
class AttrsDerived(AttrsBase):
    baz: float


@attrs.define
class AttrsBookWithMetadata:
    title: str = attrs.field(metadata={"description": "Book title"})
    author: str = attrs.field(metadata={"max_length": 100})


@attrs.define
class AttrsPersonWithLiterals:
    username: str
    role: Literal["admin", "user", "guest"]
    status: Literal["active", "inactive", "pending"]


@attrs.define
class AttrsEventWithTimeMetadata:
    """Attrs class with datetime fields that have time metadata."""

    name: str
    created_at: datetime
    scheduled_at: datetime = attrs.field(metadata={"anyschema": {"time_zone": "UTC"}})
    started_at: datetime = attrs.field(metadata={"anyschema": {"time_unit": "ms"}})
    completed_at: datetime = attrs.field(metadata={"anyschema": {"time_zone": "Europe/Berlin", "time_unit": "ns"}})


@attrs.define
class AttrsEventWithXAnyschema:
    """Attrs class using x-anyschema prefix (OpenAPI compatibility)."""

    name: str
    created_at: datetime = attrs.field(metadata={"x-anyschema": {"time_zone": "UTC"}})
    started_at: datetime = attrs.field(metadata={"x-anyschema": {"time_unit": "ms"}})


class PydanticZipcode(BaseModel):
    zipcode: PositiveInt


@attrs.define
class AttrsAddressWithPydantic:
    street: str
    city: str
    zipcode: PydanticZipcode


@dataclass
class DataclassEventWithTimeMetadata:
    """Dataclass with datetime fields that have time metadata."""

    name: str = field(metadata={"anyschema": {"description": "Event name"}})
    created_at: datetime
    scheduled_at: datetime = field(metadata={"anyschema": {"time_zone": "UTC", "description": "Scheduled time"}})
    started_at: datetime = field(metadata={"anyschema": {"time_unit": "ms"}})
    completed_at: datetime = field(metadata={"anyschema": {"time_zone": "Europe/Berlin", "time_unit": "ns"}})


def create_missing_decorator_test_case() -> tuple[type, str]:
    """Create a test case for missing decorator inheritance issue.

    Returns:
        A tuple of (child_class, expected_error_message) for testing.
    """

    @attrs.define
    class Base:
        foo: str

    class ChildWithoutDecorator(Base):
        bar: int

    expected_msg = (
        "Class 'ChildWithoutDecorator' has annotations ('bar') that are not attrs fields. "
        "If this class inherits from an attrs class, you must also decorate it with @attrs.define "
        "or @attrs.frozen to properly define these fields."
    )

    return ChildWithoutDecorator, expected_msg


class SQLAlchemyBase(DeclarativeBase):
    pass


class SimpleUserORM(SQLAlchemyBase):
    __tablename__ = "simple_user"
    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str | None]


class UserWithTypesORM(SQLAlchemyBase):
    __tablename__ = "user_with_types"
    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(String(50))
    age: Mapped[int] = mapped_column(Integer, nullable=True)
    score: Mapped[float | None]


class ProductORM(SQLAlchemyBase):
    """ORM model with multiple field types for testing."""

    __tablename__ = "product_orm"
    id: Mapped[int] = mapped_column(primary_key=True, unique=True)
    name: Mapped[str]
    price: Mapped[float]
    in_stock: Mapped[bool]


class ComplexORM(SQLAlchemyBase):
    """ORM model with various column types for testing."""

    __tablename__ = "complex_orm"
    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(String(50))
    description: Mapped[str]
    age: Mapped[int]
    score: Mapped[float] = mapped_column(Float)
    is_active: Mapped[bool]
    created_at: Mapped[DateTime] = mapped_column(DateTime)
    birth_date: Mapped[Date] = mapped_column(Date)


# Core Table instances
sql_metadata = MetaData()

user_table = Table(
    "user",
    sql_metadata,
    Column[int]("id", Integer, primary_key=True, nullable=False, doc="Primary key"),
    Column[str]("name", String(50)),
    Column[int]("age", Integer, doc="User age"),
    Column[str]("email", String(100), nullable=True),
)

numeric_table = Table(
    "numeric_table",
    sql_metadata,
    Column[int]("int_col", Integer),
    Column[int]("bigint_col", BigInteger),
    Column[str]("string_col", String(100)),
    Column[float]("float_col", Float),
)

complex_table = Table(
    "complex_table",
    sql_metadata,
    Column[int]("id", Integer, primary_key=True),
    Column[str]("name", String(50)),
    Column[str]("description", String),
    Column[int]("age", Integer),
    Column[float]("score", Float),
    Column[bool]("is_active", Boolean),
    Column[datetime]("created_at", DateTime),
    Column[date]("birth_date", Date),
)

bigint_table = Table(
    "bigint_table",
    sql_metadata,
    Column[int]("id", BigInteger, primary_key=True),
    Column[int]("count", BigInteger),
)

array_list_table = Table(
    "array_list_table",
    sql_metadata,
    Column[int]("id", Integer, primary_key=True),
    Column[Sequence[str]]("tags", ARRAY(String)),
    Column[Sequence[float]]("scores", ARRAY(Float())),
)

array_fixed_table = Table(
    "array_fixed_table",
    sql_metadata,
    Column[int]("id", Integer, primary_key=True, unique=True),
    Column[Sequence[float]]("coordinates", ARRAY(Float(), dimensions=3)),
    Column[Sequence[int]]("matrix", ARRAY(Integer, dimensions=2)),
)

# Tables with datetime metadata
event_table_with_time_metadata = Table(
    "event_with_time_metadata",
    sql_metadata,
    Column[int]("id", Integer, primary_key=True),
    Column[str]("name", String(100)),
    Column[datetime]("created_at", DateTime),  # No metadata
    Column[datetime]("scheduled_at", DateTime(timezone=True), info={"anyschema": {"time_zone": "UTC"}}),
    Column[datetime]("started_at", DateTime, info={"anyschema": {"time_unit": "ms"}}),
    Column[datetime](
        "completed_at",
        DateTime(timezone=True),
        info={"anyschema": {"time_zone": "Europe/Berlin", "time_unit": "ns"}},
    ),
)


# ORM model with datetime metadata
class EventORMWithTimeMetadata(SQLAlchemyBase):
    """ORM model with datetime fields that have time metadata."""

    __tablename__ = "event_orm_with_time_metadata"
    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str]
    created_at: Mapped[DateTime] = mapped_column(DateTime)
    scheduled_at: Mapped[DateTime] = mapped_column(DateTime(timezone=True), info={"anyschema": {"time_zone": "UTC"}})
    started_at: Mapped[DateTime] = mapped_column(DateTime, info={"anyschema": {"time_unit": "ms"}})
    completed_at: Mapped[DateTime] = mapped_column(
        DateTime(timezone=True), info={"anyschema": {"time_zone": "Europe/Berlin", "time_unit": "ns"}}
    )


# Table with timezone-aware datetime (timezone=True)
event_table_with_tz_aware = Table(
    "event_with_tz_aware",
    sql_metadata,
    Column("id", Integer, primary_key=True),
    Column(
        "timestamp_utc",
        DateTime(timezone=True),
        info={"anyschema": {"time_zone": "UTC"}},
    ),
    Column(
        "timestamp_berlin",
        DateTime(timezone=True),
        info={"anyschema": {"time_zone": "Europe/Berlin", "time_unit": "ms"}},
    ),
)



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from anyschema import AnySchema
from tests.conftest import ProductORM, array_fixed_table

if TYPE_CHECKING:
    from anyschema.typing import Spec


@pytest.mark.parametrize(
    ("spec", "expected"),
    [
        (array_fixed_table, (True, False, False)),
        (ProductORM, (True, False, False, False)),
    ],
)
def test_uniques_named_false(spec: Spec, expected: tuple[bool, ...]) -> None:
    schema = AnySchema(spec=spec)
    result = schema.uniques(named=False)

    assert result == expected


@pytest.mark.parametrize(
    ("spec", "expected"),
    [
        (array_fixed_table, {"id": True, "coordinates": False, "matrix": False}),
        (
            ProductORM,
            {"id": True, "name": False, "price": False, "in_stock": False},
        ),
    ],
)
def test_uniques_named_true(spec: Spec, expected: dict[str, bool]) -> None:
    schema = AnySchema(spec=spec)
    result = schema.uniques(named=True)

    assert result == expected



from __future__ import annotations

from datetime import date, datetime
from typing import Any

import narwhals as nw
import pytest
from pydantic import AwareDatetime, BaseModel, FutureDate, FutureDatetime, NaiveDatetime, PastDate, PastDatetime

from anyschema.exceptions import UnsupportedDTypeError
from anyschema.parsers import ParserPipeline
from anyschema.parsers._builtin import PyTypeStep
from anyschema.parsers.pydantic import PydanticTypeStep


@pytest.fixture(scope="module")
def pydantic_parser() -> PydanticTypeStep:
    """Create a PydanticTypeStep instance with pipeline set."""
    parser = PydanticTypeStep()
    py_parser = PyTypeStep()
    _ = ParserPipeline([parser, py_parser])
    return parser


@pytest.mark.parametrize(
    ("input_type", "expected"),
    [
        (NaiveDatetime, nw.Datetime()),
        (PastDatetime, nw.Datetime()),
        (FutureDatetime, nw.Datetime()),
        (PastDate, nw.Date()),
        (FutureDate, nw.Date()),
    ],
)
def test_parse_pydantic_types(pydantic_parser: PydanticTypeStep, input_type: type, expected: nw.dtypes.DType) -> None:
    result = pydantic_parser.parse(input_type, constraints=(), metadata={})
    assert result == expected


@pytest.mark.parametrize(
    ("input_type", "metadata", "expected"),
    [
        (NaiveDatetime, {"anyschema": {"time_unit": "ms"}}, nw.Datetime("ms")),
        (NaiveDatetime, {"anyschema": {"time_unit": "ns"}}, nw.Datetime("ns")),
        (PastDatetime, {"anyschema": {"time_unit": "ms"}}, nw.Datetime("ms")),
        (PastDatetime, {"anyschema": {"time_zone": "UTC"}}, nw.Datetime("us", time_zone="UTC")),
        (
            PastDatetime,
            {"anyschema": {"time_unit": "ms", "time_zone": "UTC"}},
            nw.Datetime("ms", time_zone="UTC"),
        ),
        (FutureDatetime, {"anyschema": {"time_unit": "ns"}}, nw.Datetime("ns")),
        (FutureDatetime, {"anyschema": {"time_zone": "Europe/Rome"}}, nw.Datetime("us", time_zone="Europe/Rome")),
        (
            FutureDatetime,
            {"anyschema": {"time_unit": "ns", "time_zone": "America/Los_Angeles"}},
            nw.Datetime("ns", time_zone="America/Los_Angeles"),
        ),
    ],
)
def test_parse_pydantic_datetime_with_metadata(
    pydantic_parser: PydanticTypeStep,
    input_type: type,
    metadata: dict[str, Any],
    expected: nw.dtypes.DType,
) -> None:
    """Test that pydantic datetime types parse correctly with anyschema/time_zone and anyschema/time_unit metadata."""
    result = pydantic_parser.parse(input_type, constraints=(), metadata=metadata)
    assert result == expected


def test_parse_aware_datetime_raises(pydantic_parser: PydanticTypeStep) -> None:
    expected_msg = "pydantic AwareDatetime does not specify a fixed timezone."
    with pytest.raises(UnsupportedDTypeError, match=expected_msg):
        pydantic_parser.parse(AwareDatetime, constraints=(), metadata={})


def test_parse_aware_datetime_with_tz(pydantic_parser: PydanticTypeStep) -> None:
    result = pydantic_parser.parse(
        AwareDatetime, constraints=(), metadata={"anyschema": {"time_zone": "Europe/Berlin"}}
    )
    assert result == nw.Datetime(time_zone="Europe/Berlin")


def test_parse_aware_datetime_with_tz_and_time_unit(pydantic_parser: PydanticTypeStep) -> None:
    result = pydantic_parser.parse(
        AwareDatetime,
        constraints=(),
        metadata={"anyschema": {"time_zone": "Europe/Berlin", "time_unit": "ms"}},
    )
    assert result == nw.Datetime(time_unit="ms", time_zone="Europe/Berlin")


def test_parse_model_into_struct(pydantic_parser: PydanticTypeStep) -> None:
    class SomeModel(BaseModel):
        past_date: PastDate
        future_date: FutureDate
        updated_at: NaiveDatetime

    result = pydantic_parser.parse(SomeModel, constraints=(), metadata={})

    expected_fields = [
        nw.Field(name="past_date", dtype=nw.Date()),
        nw.Field(name="future_date", dtype=nw.Date()),
        nw.Field(name="updated_at", dtype=nw.Datetime()),
    ]
    expected = nw.Struct(expected_fields)
    assert result == expected


def test_parse_nested_model(pydantic_parser: PydanticTypeStep) -> None:
    class Address(BaseModel):
        street: str
        city: str

    class Person(BaseModel):
        name: str
        address: Address

    result = pydantic_parser.parse(Person, constraints=(), metadata={})

    address_fields = [
        nw.Field(name="street", dtype=nw.String()),
        nw.Field(name="city", dtype=nw.String()),
    ]
    expected_fields = [
        nw.Field(name="name", dtype=nw.String()),
        nw.Field(name="address", dtype=nw.Struct(address_fields)),
    ]
    expected = nw.Struct(expected_fields)

    assert result == expected


def test_parse_empty_model(pydantic_parser: PydanticTypeStep) -> None:
    """Test parsing an empty Pydantic model."""

    class EmptyModel(BaseModel):
        pass

    result = pydantic_parser.parse(EmptyModel, constraints=(), metadata={})

    expected = nw.Struct([])
    assert result == expected


@pytest.mark.parametrize("input_type", [int, float, list[int], date, datetime])
def test_parse_non_pydantic_types(pydantic_parser: PydanticTypeStep, input_type: Any) -> None:
    result = pydantic_parser.parse(input_type, constraints=(), metadata={})
    assert result is None


def test_parse_custom_class_returns_none(pydantic_parser: PydanticTypeStep) -> None:
    """Test that parsing non-BaseModel class returns None."""

    class CustomClass:
        pass

    result = pydantic_parser.parse(CustomClass, constraints=(), metadata={})
    assert result is None


def test_parse_model_with_field_metadata(pydantic_parser: PydanticTypeStep) -> None:
    """Test parsing model that has field metadata."""
    from typing import Annotated

    from pydantic import Field

    class ModelWithConstraints(BaseModel):
        name: Annotated[str, Field(min_length=1, max_length=100)]
        age: Annotated[int, Field(gt=0, lt=150)]

    result = pydantic_parser.parse(ModelWithConstraints, constraints=(), metadata={})

    # The metadata is stored in field_info.metadata but the parsing should still work
    expected_fields = [
        nw.Field(name="name", dtype=nw.String()),
        nw.Field(name="age", dtype=nw.Int64()),
    ]
    expected = nw.Struct(expected_fields)

    assert result == expected


def test_parse_naive_datetime_with_timezone_raises(pydantic_parser: PydanticTypeStep) -> None:
    """Test that NaiveDatetime with timezone raises an error."""
    expected_msg = "pydantic NaiveDatetime should not specify a timezone, found UTC."
    with pytest.raises(UnsupportedDTypeError, match=expected_msg):
        pydantic_parser.parse(NaiveDatetime, constraints=(), metadata={"anyschema": {"time_zone": "UTC"}})



from __future__ import annotations

from typing import TYPE_CHECKING, Mapping

import narwhals as nw
import pytest

from anyschema import AnySchema
from tests.conftest import (
    ComplexORM,
    EventORMWithTimeMetadata,
    ProductORM,
    SimpleUserORM,
    array_fixed_table,
    array_list_table,
    bigint_table,
    complex_table,
    event_table_with_time_metadata,
    event_table_with_tz_aware,
    user_table,
)

if TYPE_CHECKING:
    from anyschema.typing import SQLAlchemyTableType


@pytest.mark.parametrize(
    ("spec", "expected_schema"),
    [
        # ORM models
        (
            SimpleUserORM,
            {
                "id": nw.Int32(),
                "name": nw.String(),
            },
        ),
        (
            ProductORM,
            {
                "id": nw.Int32(),
                "name": nw.String(),
                "price": nw.Float32(),
                "in_stock": nw.Boolean(),
            },
        ),
        (
            ComplexORM,
            {
                "id": nw.Int32(),
                "name": nw.String(),
                "description": nw.String(),
                "age": nw.Int32(),
                "score": nw.Float32(),
                "is_active": nw.Boolean(),
                "created_at": nw.Datetime(),
                "birth_date": nw.Date(),
            },
        ),
        # Core tables
        (
            user_table,
            {
                "id": nw.Int32(),
                "name": nw.String(),
                "age": nw.Int32(),
                "email": nw.String(),
            },
        ),
        (
            complex_table,
            {
                "id": nw.Int32(),
                "name": nw.String(),
                "description": nw.String(),
                "age": nw.Int32(),
                "score": nw.Float32(),
                "is_active": nw.Boolean(),
                "created_at": nw.Datetime(),
                "birth_date": nw.Date(),
            },
        ),
        (
            bigint_table,
            {
                "id": nw.Int64(),
                "count": nw.Int64(),
            },
        ),
        # Array types - List (no dimensions)
        (
            array_list_table,
            {
                "id": nw.Int32(),
                "tags": nw.List(nw.String()),
                "scores": nw.List(nw.Float32()),
            },
        ),
        # Array types - Fixed dimensions (Array)
        (
            array_fixed_table,
            {
                "id": nw.Int32(),
                "coordinates": nw.Array(nw.Float32(), shape=(3,)),
                "matrix": nw.Array(nw.Int32(), shape=(2,)),
            },
        ),
    ],
)
def test_sqlalchemy_spec(spec: SQLAlchemyTableType, expected_schema: Mapping[str, nw.dtypes.DType]) -> None:
    schema = AnySchema(spec=spec)
    nw_schema = schema._nw_schema
    assert nw_schema == nw.Schema(expected_schema)


@pytest.mark.parametrize(
    ("spec", "expected_schema"),
    [
        # Table with time metadata
        (
            event_table_with_time_metadata,
            {
                "id": nw.Int32(),
                "name": nw.String(),
                "created_at": nw.Datetime(),
                "scheduled_at": nw.Datetime(time_zone="UTC"),
                "started_at": nw.Datetime(time_unit="ms"),
                "completed_at": nw.Datetime(time_unit="ns", time_zone="Europe/Berlin"),
            },
        ),
        # ORM with time metadata
        (
            EventORMWithTimeMetadata,
            {
                "id": nw.Int32(),
                "name": nw.String(),
                "created_at": nw.Datetime(),
                "scheduled_at": nw.Datetime(time_zone="UTC"),
                "started_at": nw.Datetime(time_unit="ms"),
                "completed_at": nw.Datetime(time_unit="ns", time_zone="Europe/Berlin"),
            },
        ),
        # Table with timezone-aware datetime
        (
            event_table_with_tz_aware,
            {
                "id": nw.Int32(),
                "timestamp_utc": nw.Datetime(time_zone="UTC"),
                "timestamp_berlin": nw.Datetime(time_unit="ms", time_zone="Europe/Berlin"),
            },
        ),
    ],
)
def test_sqlalchemy_spec_with_time_metadata(
    spec: SQLAlchemyTableType, expected_schema: Mapping[str, nw.dtypes.DType]
) -> None:
    """Test that SQLAlchemy specs with time metadata are correctly converted to narwhals schema."""
    schema = AnySchema(spec=spec)
    nw_schema = schema._nw_schema
    assert nw_schema == nw.Schema(expected_schema)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pytest

from anyschema import AnySchema

if TYPE_CHECKING:
    from anyschema.typing import Spec


@pytest.mark.parametrize(
    "spec",
    [
        {"name": str, "age": int},
        {"users": list[str], "counts": dict[str, int]},
    ],
)
def test_equal(spec: Spec) -> None:
    assert AnySchema(spec=spec) == AnySchema(spec=spec)


@pytest.mark.parametrize(
    ("spec1", "spec2"),
    [
        ({"name": str, "age": int}, {"age": int, "name": str}),
        ({"name": str, "age": int}, {"name": str}),
        ({"value": int}, {"value": float}),
        ({"name": str}, {"name": str | None}),
    ],
)
def test_not_equal(spec1: Spec, spec2: Spec) -> None:
    schema1 = AnySchema(spec=spec1)
    schema2 = AnySchema(spec=spec2)

    assert schema1 != schema2


@pytest.mark.parametrize(
    "other",
    [
        {"name": str},
        "not a schema",
        42,
        None,
    ],
)
def test_equality_with_non_anyschema_object(other: Any) -> None:
    schema = AnySchema(spec={"name": str})

    assert schema != other



from __future__ import annotations

from typing import TYPE_CHECKING, Annotated

import narwhals as nw
from annotated_types import Interval
from pydantic import BaseModel

from tests.pydantic.utils import model_to_nw_schema

if TYPE_CHECKING:
    from anyschema.parsers import ParserPipeline


def test_parse_struct(auto_pipeline: ParserPipeline) -> None:
    class BaseStruct(BaseModel):
        x1: Annotated[int, Interval(gt=0, lt=123)]
        x2: str
        x3: float | None
        x4: None | bool

    class StructModel(BaseModel):
        struct: BaseStruct | None

    schema = model_to_nw_schema(StructModel, pipeline=auto_pipeline)
    expected = {
        "struct": nw.Struct(
            [
                nw.Field("x1", nw.UInt8()),
                nw.Field("x2", nw.String()),
                nw.Field("x3", nw.Float64()),
                nw.Field("x4", nw.Boolean()),
            ]
        )
    }
    assert schema == expected



from __future__ import annotations

from typing import TYPE_CHECKING

import narwhals as nw
from pydantic import BaseModel, StrictBool

from tests.pydantic.utils import model_to_nw_schema

if TYPE_CHECKING:
    from anyschema.parsers import ParserPipeline


def test_parse_boolean(auto_pipeline: ParserPipeline) -> None:
    class BooleanModel(BaseModel):
        # python bool type
        py_bool: bool
        py_bool_optional: bool | None
        py_bool_or_none: bool | None
        none_or_py_bool: None | bool

        # pydantic StrictBool type
        strict_bool: StrictBool
        strict_bool_optional: StrictBool | None
        strict_bool_or_none: StrictBool | None
        none_or_strict_bool: None | StrictBool

    schema = model_to_nw_schema(BooleanModel, pipeline=auto_pipeline)

    assert all(value == nw.Boolean() for value in schema.values())



from __future__ import annotations

from collections.abc import Callable, Generator, Mapping, Sequence
from typing import TYPE_CHECKING, Annotated, Any, Literal, Protocol, TypeAlias, TypedDict

if TYPE_CHECKING:
    from dataclasses import Field as DataclassField
    from typing import ClassVar

    from attrs import AttrsInstance
    from narwhals.dtypes import DType
    from narwhals.schema import Schema
    from narwhals.typing import TimeUnit
    from pydantic import BaseModel
    from sqlalchemy import Table
    from sqlalchemy.orm import DeclarativeBase
    from sqlalchemy.sql.type_api import TypeEngine

    from anyschema.parsers import ParserStep

    AttrsClassType: TypeAlias = type[AttrsInstance]
    SQLAlchemyTableType: TypeAlias = Table | type[DeclarativeBase]


IntoOrderedDict: TypeAlias = Mapping[str, type] | Sequence[tuple[str, type]]
"""An object that can be converted into a python [`OrderedDict`][ordered-dict].

We check for the object to be either a mapping or a sequence of sized 2 tuples.

[ordered-dict]: https://docs.python.org/3/library/collections.html#collections.OrderedDict
"""

IntoParserPipeline: TypeAlias = "Literal['auto'] | Sequence['ParserStep']"
"""An object that can be converted into a [`ParserPipeline`][anyschema.parsers.ParserPipeline].

Either "auto" or a sequence of [`ParserStep`][anyschema.parsers.ParserStep].
"""

UnknownSpec: TypeAlias = Any
"""An unknown specification."""

Spec: TypeAlias = "Schema | IntoOrderedDict | type[BaseModel] | DataclassType | TypedDictType | AttrsClassType | SQLAlchemyTableType | UnknownSpec"  # noqa: E501
"""Input specification supported directly by [`AnySchema`][anyschema.AnySchema]."""

FieldName: TypeAlias = str
FieldType: TypeAlias = "type[Any] | Annotated[Any, ...] | TypeEngine[Any]"
FieldConstraints: TypeAlias = tuple[Any, ...]
FieldMetadata: TypeAlias = dict[str, Any]

FieldSpec: TypeAlias = tuple[FieldName, FieldType, FieldConstraints, FieldMetadata]
"""Field specification: alias for a tuple of `(str, type, tuple(constraints, ...), dict(metadata))`."""

FieldSpecIterable: TypeAlias = Generator[FieldSpec, None, None]
"""Return type of an adapter."""

Adapter: TypeAlias = Callable[[Any], FieldSpecIterable]
"""Adapter expected signature.

An adapter is a callable that adapts a spec into field specifications.
"""


class DataclassInstance(Protocol):
    """Protocol that represents a dataclass in Python."""

    # dataclasses are runtime composed entities making them tricky to type, this may not work perfectly
    #   with all type checkers
    # code adapted from typeshed:
    # https://github.com/python/typeshed/blob/9ab7fde0a0cd24ed7a72837fcb21093b811b80d8/stdlib/_typeshed/__init__.pyi#L351
    __dataclass_fields__: ClassVar[dict[str, DataclassField[Any]]]


DataclassType = type[DataclassInstance]


class TypedDictType(Protocol):
    """Protocol that represents a TypedDict in Python."""

    __annotations__: dict[str, type]
    __required_keys__: frozenset[str]
    __optional_keys__: frozenset[str]


class AnySchemaMetadata(TypedDict, total=False):
    """TypedDict for anyschema-specific metadata keys.

    This structure defines the nested metadata format that anyschema recognizes
    for controlling field parsing behavior. All keys are optional.

    Attributes:
        description: Human-readable description of the field.
        dtype: Narwhals DType (or its serialized/string representation)
        nullable: Whether the field can contain null values.
        time_zone: Timezone for datetime fields (e.g., "UTC", "Europe/Berlin").
        time_unit: Time precision for datetime fields ("s", "ms", "us", "ns").
        unique: Whether all values in the field must be unique.

    Examples:
        >>> metadata: AnySchemaMetadata = {"nullable": True, "time_zone": "UTC"}
        >>> metadata["unique"] = False
    """

    description: str | None
    dtype: str | DType
    nullable: bool
    time_zone: str
    time_unit: TimeUnit
    unique: bool


AnySchemaMetadataKey: TypeAlias = Literal["description", "dtype", "nullable", "time_zone", "time_unit", "unique"]
AnySchemaNamespaceKey: TypeAlias = Literal["anyschema", "x-anyschema"]



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from typing_extensions import Any, assert_type

import anyschema


def test_package_getattr() -> None:
    assert_type(anyschema.__version__, str)
    assert_type(anyschema.__title__, str)

    if TYPE_CHECKING:
        bad = anyschema.not_real  # type: ignore[attr-defined]
        assert_type(bad, Any)

    with pytest.raises(AttributeError):
        very_bad = anyschema.not_real  # type: ignore[attr-defined]  # noqa: F841



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from anyschema.adapters import into_ordered_dict_adapter

if TYPE_CHECKING:
    from anyschema.typing import FieldSpec, IntoOrderedDict


@pytest.mark.parametrize(
    "spec",
    [
        {"name": str, "age": int},
        [("name", str), ("age", int)],
    ],
)
def test_into_ordered_dict_adapter(spec: IntoOrderedDict) -> None:
    expected: tuple[FieldSpec, ...] = (("name", str, (), {}), ("age", int, (), {}))
    result = tuple(into_ordered_dict_adapter(spec))
    assert result == expected



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ForwardRef

import narwhals as nw
import pytest

from anyschema.parsers import ForwardRefStep, ParserPipeline, PyTypeStep, UnionTypeStep

if TYPE_CHECKING:
    from contextlib import AbstractContextManager


def forward_ref_parser_builder(
    globalns: dict[str, Any] | None = None, localns: dict[str, Any] | None = None
) -> ForwardRefStep:
    forward_ref_parser = ForwardRefStep(globalns=globalns, localns=localns)
    union_parser = UnionTypeStep()
    py_parser = PyTypeStep()
    _ = ParserPipeline([forward_ref_parser, union_parser, py_parser])
    return forward_ref_parser


@pytest.fixture(scope="module")
def forward_ref_parser() -> ForwardRefStep:
    """Create a ForwardRefStep instance with pipeline set."""
    return forward_ref_parser_builder()


class CustomType:
    pass


@pytest.mark.parametrize(
    ("input_type", "expected"),
    [
        ("int", nw.Int64()),
        ("str", nw.String()),
        ("float", nw.Float64()),
        ("bool", nw.Boolean()),
        ("list[int]", nw.List(nw.Int64())),
        ("list[str]", nw.List(nw.String())),
        ("tuple[int, ...]", nw.List(nw.Int64())),
        ("List[int]", nw.List(nw.Int64())),
        ("int | None", nw.Int64()),
        ("Optional[int]", nw.Int64()),
    ],
)
def test_forward_ref(forward_ref_parser: ForwardRefStep, input_type: Any, expected: nw.dtypes.DType) -> None:
    result = forward_ref_parser.parse(input_type=ForwardRef(input_type), constraints=(), metadata={})
    assert result == expected


@pytest.mark.parametrize(
    "input_type",
    [
        int,
        str,
        list[int],
    ],
)
def test_non_forward_ref(forward_ref_parser: ForwardRefStep, input_type: Any) -> None:
    result = forward_ref_parser.parse(input_type=input_type, constraints=(), metadata={})
    assert result is None


def test_custom_globals() -> None:
    custom_globals = {
        "CustomType": int,  # Custom class to a parsable type
        "str": bool,  # don't do this in production
    }
    parser = forward_ref_parser_builder(globalns=custom_globals)
    result = parser.parse(ForwardRef("CustomType"), constraints=(), metadata={})
    assert result == nw.Int64()

    result = parser.parse(ForwardRef("str"), constraints=(), metadata={})
    assert result == nw.Boolean()


def test_custom_local() -> None:
    class LocalClass:  # pyright: ignore[reportUnusedClass]
        pass

    parser = forward_ref_parser_builder(localns={"LocalClass": int})
    result = parser.parse(ForwardRef("LocalClass"), constraints=(), metadata={})
    assert result == nw.Int64()


@pytest.mark.parametrize(
    ("input_type", "context"),
    [
        ("UndefinedType", pytest.raises(NotImplementedError, match="Failed to resolve ForwardRef")),
        ("list[", pytest.raises(SyntaxError, match="Forward reference must be an expression")),
        ("1 + 1", pytest.raises(NotImplementedError, match="No parser in the pipeline could handle type")),
    ],
)
def test_resolution_error(
    forward_ref_parser: ForwardRefStep, input_type: str, context: AbstractContextManager[Any]
) -> None:
    with context:
        forward_ref_parser.parse(ForwardRef(input_type), constraints=(), metadata={})


def test_build_namespace_default(forward_ref_parser: ForwardRefStep) -> None:
    # Globals
    assert forward_ref_parser.globalns["int"] is int
    assert forward_ref_parser.globalns["str"] is str
    assert forward_ref_parser.globalns["float"] is float
    assert forward_ref_parser.globalns["bool"] is bool
    assert forward_ref_parser.globalns["list"] is list
    assert "Optional" in forward_ref_parser.globalns
    assert "Union" in forward_ref_parser.globalns
    assert "List" in forward_ref_parser.globalns

    # Locals
    assert forward_ref_parser.localns == {}


def test_build_namespace_custom() -> None:
    custom_globals = {"CustomType": CustomType, "MyInt": int, "bool": str}
    custom_locals = {"local_var": int}
    parser = forward_ref_parser_builder(globalns=custom_globals, localns=custom_locals)

    # Check that custom types are in the namespace
    assert parser.globalns["CustomType"] is CustomType
    assert parser.globalns["MyInt"] is int
    # Built-ins should still be there
    assert parser.globalns["int"] is int
    assert parser.globalns["str"] is str

    assert parser.globalns["bool"] is str

    assert parser.localns == custom_locals


def test_build_namespace_includes_pydantic_types() -> None:
    from anyschema._dependencies import PYDANTIC_AVAILABLE

    parser = forward_ref_parser_builder()

    if PYDANTIC_AVAILABLE:
        assert "BaseModel" in parser.globalns
        assert "Field" in parser.globalns
        assert "PositiveInt" in parser.globalns
        assert "NegativeInt" in parser.globalns
        assert "NonPositiveInt" in parser.globalns
        assert "NonNegativeInt" in parser.globalns
        assert "PositiveFloat" in parser.globalns
        assert "NegativeFloat" in parser.globalns
        assert "NonPositiveFloat" in parser.globalns
        assert "NonNegativeFloat" in parser.globalns
        assert "constr" in parser.globalns
        assert "conint" in parser.globalns
        assert "confloat" in parser.globalns
        assert "conlist" in parser.globalns
        assert "conset" in parser.globalns
    else:  # pragma: no cover
        ...


def test_build_namespace_includes_annotated_types() -> None:
    from anyschema._dependencies import ANNOTATED_TYPES_AVAILABLE

    parser = forward_ref_parser_builder()

    if ANNOTATED_TYPES_AVAILABLE:
        # Check that annotated_types are in the namespace
        assert "Gt" in parser.globalns
        assert "Ge" in parser.globalns
        assert "Lt" in parser.globalns
        assert "Le" in parser.globalns
        assert "Interval" in parser.globalns
        assert "MultipleOf" in parser.globalns
        assert "MinLen" in parser.globalns
        assert "MaxLen" in parser.globalns
        assert "Len" in parser.globalns
        assert "Timezone" in parser.globalns
        assert "Predicate" in parser.globalns
    else:  # pragma: no cover
        ...



from __future__ import annotations

__all__ = ("UnavailablePipelineError", "UnsupportedDTypeError")


class UnavailablePipelineError(ValueError):
    """Exception raised when a parser does not have a ParserPipeline set."""


class UnsupportedDTypeError(ValueError):
    """Exception raised when a DType is not supported."""



from __future__ import annotations

import typing as _t

from anyschema._anyschema import AnyField, AnySchema
from anyschema._utils import show_versions

__all__ = ("AnyField", "AnySchema", "show_versions")
__title__ = __name__
__version__: str


if not _t.TYPE_CHECKING:

    def __getattr__(name: str) -> _t.Any:
        if name == "__version__":
            global __version__  # noqa: PLW0603

            from importlib import metadata

            __version__ = metadata.version(__name__)
            return __version__
        msg = f"module {__name__!r} has no attribute {name!r}"
        raise AttributeError(msg)
else:  # pragma: no cover
    ...



from __future__ import annotations

import attrs
import narwhals as nw
import pytest

from anyschema.parsers import ParserPipeline, PyTypeStep
from anyschema.parsers.attrs import AttrsTypeStep
from tests.conftest import AttrsDerived, AttrsPerson, AttrsPersonFrozen, create_missing_decorator_test_case


@pytest.fixture(scope="module")
def attrs_parser() -> AttrsTypeStep:
    attrs_parser = AttrsTypeStep()
    py_parser = PyTypeStep()
    _ = ParserPipeline([attrs_parser, py_parser])
    return attrs_parser


def test_parse_attrs_class_into_struct(attrs_parser: AttrsTypeStep) -> None:
    result = attrs_parser.parse(AttrsPerson, (), {})

    expected_fields = [
        nw.Field(name="name", dtype=nw.String()),
        nw.Field(name="age", dtype=nw.Int64()),
        nw.Field(name="date_of_birth", dtype=nw.Date()),
        nw.Field(name="is_active", dtype=nw.Boolean()),
        nw.Field(name="classes", dtype=nw.List(nw.String())),
        nw.Field(name="grades", dtype=nw.List(nw.Float64())),
    ]
    expected = nw.Struct(expected_fields)
    assert result == expected


def test_parse_frozen_attrs_class(attrs_parser: AttrsTypeStep) -> None:
    result = attrs_parser.parse(AttrsPersonFrozen, (), {})

    expected_fields = [
        nw.Field(name="name", dtype=nw.String()),
        nw.Field(name="age", dtype=nw.Int64()),
        nw.Field(name="date_of_birth", dtype=nw.Date()),
    ]
    expected = nw.Struct(expected_fields)
    assert result == expected


def test_parse_empty_attrs_class(attrs_parser: AttrsTypeStep) -> None:
    @attrs.define
    class EmptyClass:
        pass

    result = attrs_parser.parse(EmptyClass, (), {})
    expected = nw.Struct([])
    assert result == expected


def test_parse_non_attrs_class_returns_none(attrs_parser: AttrsTypeStep) -> None:
    class RegularClass:
        pass

    result = attrs_parser.parse(RegularClass, (), {})
    assert result is None


def test_parse_classic_attr_s_decorator(attrs_parser: AttrsTypeStep) -> None:
    import attr

    @attr.s(auto_attribs=True)
    class ClassicAttrs:
        name: str
        value: int

    result = attrs_parser.parse(ClassicAttrs, (), {})

    expected_fields = [
        nw.Field(name="name", dtype=nw.String()),
        nw.Field(name="value", dtype=nw.Int64()),
    ]
    expected = nw.Struct(expected_fields)
    assert result == expected


def test_parse_attrs_with_inheritance(attrs_parser: AttrsTypeStep) -> None:
    result = attrs_parser.parse(AttrsDerived, (), {})

    expected_fields = [
        nw.Field(name="foo", dtype=nw.String()),
        nw.Field(name="bar", dtype=nw.Int64()),
        nw.Field(name="baz", dtype=nw.Float64()),
    ]
    expected = nw.Struct(expected_fields)
    assert result == expected


def test_parse_attrs_missing_decorator_raises(attrs_parser: AttrsTypeStep) -> None:
    child_cls, expected_msg = create_missing_decorator_test_case()
    with pytest.raises(AssertionError, match=expected_msg.replace("(", r"\(").replace(")", r"\)")):
        attrs_parser.parse(child_cls, (), {})



from __future__ import annotations

from collections.abc import Iterable, Mapping, Sequence
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from enum import Enum
from typing import TYPE_CHECKING, Any, Literal

import narwhals as nw
from typing_extensions import get_args, get_origin, get_type_hints  # noqa: UP035

from anyschema._dependencies import is_typed_dict
from anyschema._metadata import get_anyschema_value_by_key
from anyschema.exceptions import UnsupportedDTypeError
from anyschema.parsers._base import ParserStep

if TYPE_CHECKING:
    from narwhals.dtypes import DType

    from anyschema.typing import FieldConstraints, FieldMetadata, FieldType, TypedDictType


class PyTypeStep(ParserStep):
    """Parser for Python builtin types.

    Handles:

    - `int`, `float`, `decimal`, `str`, `bytes`, `bool`, `date`, `datetime`, `timedelta`, `time`, `object`, `Enum`,
        `Literal`
    - generics: `list[T]`, `Sequence[T]`, `Iterable[T]`, `tuple[T, ...]`
    - `dict`,` Mapping[K, V]`, and typed dictionaries (`TypedDict`)
    """

    def parse(self, input_type: FieldType, constraints: FieldConstraints, metadata: FieldMetadata) -> DType | None:  # noqa: C901, PLR0912
        """Parse Python type annotations into Narwhals dtypes.

        Arguments:
            input_type: The type to parse.
            constraints: Constraints associated with the type.
            metadata: Custom metadata dictionary.

        Returns:
            A Narwhals DType if this parser can handle the type, None otherwise.
        """
        # Handle generic types first: list[T], tuple[T, ...], Sequence[T], Iterable[T], dict[K, V], Literal[...]
        # Note: In Python 3.10, generic aliases like list[int] pass isinstance(input_type, type),
        # but they cannot be used with issubclass() against abstract base classes like Sequence/Iterable.
        # Checking get_origin() first avoids this issue.
        if (origin := get_origin(input_type)) is not None:
            return self._parse_generic(input_type, origin, constraints, metadata)

        # Now handle actual classes (not generic aliases)
        if isinstance(input_type, type):
            # NOTE: The order is quite important. In fact:
            #   * issubclass(MyEnum(str, Enum), str) -> True
            #   * issubclass(bool, int) -> True
            #   * issubclass(TypedDict, dict) -> True
            #   * issubclass(datetime, date) -> True
            if issubclass(input_type, Enum):
                return nw.Enum(input_type)
            if issubclass(input_type, str):
                return nw.String()
            if issubclass(input_type, bool):
                return nw.Boolean()
            if issubclass(input_type, int):
                return nw.Int64()
            if issubclass(input_type, float):
                return nw.Float64()
            if issubclass(input_type, datetime):
                return nw.Datetime(
                    time_unit=get_anyschema_value_by_key(metadata, key="time_unit", default="us"),
                    time_zone=get_anyschema_value_by_key(metadata, key="time_zone"),
                )
            if issubclass(input_type, date):
                return nw.Date()
            if issubclass(input_type, timedelta):
                return nw.Duration()
            if issubclass(input_type, time):
                return nw.Time()
            if issubclass(input_type, Decimal):
                return nw.Decimal()
            if issubclass(input_type, bytes):
                return nw.Binary()
            if is_typed_dict(input_type):
                return self._parse_typed_dict(input_type, constraints, metadata)
            if issubclass(input_type, dict):  # Plain dict without type parameters -> Struct with Object fields
                return nw.Struct([])
            # TODO(FBruzzesi): https://github.com/FBruzzesi/anyschema/issues/56
            if issubclass(input_type, (set, frozenset)):
                return None
            if issubclass(input_type, (list, tuple, Sequence, Iterable)):
                return nw.List(nw.Object())

        if input_type is object:
            return nw.Object()

        return None

    def _parse_generic(
        self, input_type: FieldType, origin: Any, constraints: FieldConstraints, metadata: FieldMetadata
    ) -> DType | None:
        """Parse generic types like list[int], dict[str, int].

        Arguments:
            input_type: The generic type to parse.
            origin: result of `get_origin(input_type)`, passed to avoid recomputing it.
            constraints: Constraints associated with the type.
            metadata: Custom metadata dictionary.

        Returns:
            A Narwhals DType if this parser can handle the type, None otherwise.
        """
        if origin is Literal:
            return nw.Enum(get_args(input_type))

        if origin in (dict, Mapping):
            # For now, we treat dict[K, V] as an empty Struct
            # TODO(FBruzzesi): What's a better way to map this? We should introspect the mapping values
            return nw.Struct([])

        args = get_args(input_type)
        inner_dtype = self.pipeline.parse(args[0], constraints=constraints, metadata=metadata, strict=True)

        if origin in (list, Sequence, Iterable):
            return nw.List(inner_dtype)

        if origin is tuple:
            if len(args) == 2 and args[1] is Ellipsis:  # noqa: PLR2004
                # tuple[T, ...] - variable length tuple
                return nw.List(inner_dtype)

            if len(set(args)) != 1:
                msg = f"Tuple with mixed types is not supported: {input_type}"
                raise UnsupportedDTypeError(msg)

            return nw.Array(inner_dtype, shape=len(args))

        return None

    def _parse_typed_dict(
        self,
        typed_dict: TypedDictType,
        constraints: FieldConstraints,  # noqa: ARG002
        metadata: FieldMetadata,  # noqa: ARG002
    ) -> DType:
        """Parse a TypedDict into a Struct type.

        Arguments:
            typed_dict: The TypedDict class.
            constraints: Constraints associated with the type.
            metadata: Custom metadata dictionary.

        Returns:
            A Narwhals Struct dtype.
        """
        try:
            type_hints = get_type_hints(typed_dict)
        except Exception:  # pragma: no cover  # noqa: BLE001
            # If we can't get type hints, use __annotations__
            type_hints = getattr(typed_dict, "__annotations__", {})

        fields = [
            nw.Field(name=field_name, dtype=self.pipeline.parse(field_type, constraints=(), metadata={}))
            for field_name, field_type in type_hints.items()
        ]
        return nw.Struct(fields)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ForwardRef

from typing_extensions import evaluate_forward_ref

from anyschema._dependencies import ANNOTATED_TYPES_AVAILABLE, PYDANTIC_AVAILABLE
from anyschema.parsers._base import ParserStep

if TYPE_CHECKING:
    from collections.abc import Mapping

    from narwhals.dtypes import DType

    from anyschema.typing import FieldConstraints, FieldMetadata, FieldType


class ForwardRefStep(ParserStep):
    """Parser for ForwardRef types (string annotations and forward references).

    This parser handles type annotations that are forward references (ForwardRef),
    which occur when using string annotations or referencing types before they're defined.

    The parser resolves the ForwardRef to the actual type and delegates to the parser chain.

    Initialization can be customized by providing a global and local namespace that will be used to evaluate the
    forward references when resolving the type. The default namespace is built with common types, yet you can override
    it with your own types.

    Arguments:
        globalns: Global namespace for evaluating forward references. Defaults to a namespace with common types.
        localns: Local namespace for evaluating forward references. Defaults to an empty namespace.
    """

    localns: Mapping[str, Any]

    def __init__(self, globalns: dict[str, Any] | None = None, localns: Mapping[str, Any] | None = None) -> None:
        super().__init__()
        # Build namespace with common types for resolution
        self.globalns = self._build_namespace(globalns)
        self.localns = localns if localns is not None else {}

    def _build_namespace(self, user_globals: dict[str, Any] | None) -> dict[str, Any]:
        """Build a namespace with common types for ForwardRef resolution.

        Arguments:
            user_globals: User-provided global namespace.

        Returns:
            A namespace with built-in types and typing constructs.
        """
        namespace = {
            # Built-in types
            "int": int,
            "str": str,
            "float": float,
            "bool": bool,
            "list": list,
            "dict": dict,
            "tuple": tuple,
            "set": set,
            # Typing constructs
            "Annotated": __import__("typing").Annotated,
            "Dict": __import__("typing").Dict,
            "List": __import__("typing").List,
            "Optional": __import__("typing").Optional,
            "Set": __import__("typing").Set,
            "Union": __import__("typing").Union,
            "Tuple": __import__("typing").Tuple,
        }

        # Add pydantic types if available
        if PYDANTIC_AVAILABLE:
            import pydantic

            pydantic_types = {
                "BaseModel": pydantic.BaseModel,
                "Field": pydantic.Field,
                # Constrained integers
                "PositiveInt": pydantic.PositiveInt,
                "NegativeInt": pydantic.NegativeInt,
                "NonPositiveInt": pydantic.NonPositiveInt,
                "NonNegativeInt": pydantic.NonNegativeInt,
                # Constrained floats
                "PositiveFloat": pydantic.PositiveFloat,
                "NegativeFloat": pydantic.NegativeFloat,
                "NonPositiveFloat": pydantic.NonPositiveFloat,
                "NonNegativeFloat": pydantic.NonNegativeFloat,
                # constraints
                "constr": pydantic.constr,
                "conint": pydantic.conint,
                "confloat": pydantic.confloat,
                "conlist": pydantic.conlist,
                "conset": pydantic.conset,
            }
            namespace.update(pydantic_types)

        # Add annotated-types if available
        if ANNOTATED_TYPES_AVAILABLE:
            import annotated_types

            annotated_types_dict = {
                "Gt": annotated_types.Gt,
                "Ge": annotated_types.Ge,
                "Lt": annotated_types.Lt,
                "Le": annotated_types.Le,
                "Interval": annotated_types.Interval,
                "MultipleOf": annotated_types.MultipleOf,
                "MinLen": annotated_types.MinLen,
                "MaxLen": annotated_types.MaxLen,
                "Len": annotated_types.Len,
                "Timezone": annotated_types.Timezone,
                "Predicate": annotated_types.Predicate,
            }
            namespace.update(annotated_types_dict)

        # Add user-provided globals (can override defaults)
        if user_globals:
            namespace.update(user_globals)

        return namespace

    def parse(self, input_type: FieldType, constraints: FieldConstraints, metadata: FieldMetadata) -> DType | None:
        """Parse ForwardRef types by resolving them and delegating to the chain.

        Arguments:
            input_type: The type to parse (may be a ForwardRef).
            constraints: Constraints associated with the type.
            metadata: Custom metadata dictionary.

        Returns:
            A Narwhals DType if this is a ForwardRef that can be resolved, None otherwise.
        """
        if not isinstance(input_type, ForwardRef):
            return None

        try:
            resolved_type = evaluate_forward_ref(
                forward_ref=input_type,
                globals=self.globalns,
                locals=self.localns,
            )
        except (NameError, AttributeError, TypeError) as e:
            # If resolution fails, we can't handle this type
            msg = f"Failed to resolve ForwardRef '{input_type.__forward_arg__}': {e}"
            raise NotImplementedError(msg) from e

        return self.pipeline.parse(resolved_type, constraints, metadata, strict=True)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeAlias

import narwhals as nw
from sqlalchemy import types as sqltypes
from typing_extensions import TypeIs

from anyschema._metadata import get_anyschema_value_by_key
from anyschema._utils import is_sequence_of
from anyschema.exceptions import UnsupportedDTypeError
from anyschema.parsers._base import ParserStep

if TYPE_CHECKING:
    from narwhals.dtypes import DType

    from anyschema.typing import FieldConstraints, FieldMetadata, FieldType

__all__ = ("SQLAlchemyTypeStep",)

STRING_TYPES = (
    sqltypes.String,  # Includes  sqltypes.Text, sqltypes.Unicode, sqltypes.UnicodeText
    sqltypes.JSON,  # In python and dataframes this is probably the best we can do for JSON
    sqltypes.Uuid,  # In python and dataframes this is probably the best we can do for UUID
)

SupportedSQLAlchemyType: TypeAlias = (
    sqltypes.String
    | sqltypes.Boolean
    | sqltypes.ARRAY[Any]
    | sqltypes.Integer
    | sqltypes.Numeric[Any]
    | sqltypes.Date
    | sqltypes.DateTime
    | sqltypes.Time
    | sqltypes.Interval
    | sqltypes._Binary  # noqa: SLF001
    | sqltypes.JSON
    | sqltypes.Uuid[Any]
)
_SUPPORTED_SQLALCHEMY_TYPES = (
    sqltypes.String,
    sqltypes.Boolean,
    sqltypes.ARRAY,
    sqltypes.Integer,
    sqltypes.Numeric,
    sqltypes.Date,
    sqltypes.DateTime,
    sqltypes.Time,
    sqltypes.Interval,
    sqltypes._Binary,  # noqa: SLF001
    sqltypes.JSON,
    sqltypes.Uuid,
)


def _is_supported_sqlalchemy_type(obj: FieldType) -> TypeIs[SupportedSQLAlchemyType]:
    return isinstance(obj, _SUPPORTED_SQLALCHEMY_TYPES)


class SQLAlchemyTypeStep(ParserStep):
    """Parser for SQLAlchemy-specific types.

    Handles:

    - SQLAlchemy column types (Integer, String, DateTime, etc.)
    - SQLAlchemy relationship types
    - SQLAlchemy custom types

    Warning:
        It requires [sqlalchemy](https://www.sqlalchemy.org/) to be installed.
    """

    def parse(  # noqa: C901, PLR0912
        self,
        input_type: FieldType,
        constraints: FieldConstraints,
        metadata: FieldMetadata,
    ) -> DType | None:
        """Parse SQLAlchemy-specific types into Narwhals dtypes.

        Arguments:
            input_type: The type to parse.
            constraints: Constraints associated with the type.
            metadata: Custom metadata dictionary.

        Returns:
            A Narwhals DType if this parser can handle the type, None otherwise.
        """
        if not isinstance(input_type, sqltypes.TypeEngine):  # Keep this as a fast path!
            return None

        if not _is_supported_sqlalchemy_type(input_type):
            return None

        # NOTE: The order is quite important. In fact:
        #   * issubclass(Enum(...), String) -> True
        #   * issubclass(SmallInteger(), Integer) -> True
        #   * issubclass(Double(), Float) -> True
        if isinstance(input_type, sqltypes.Enum):
            categories = input_type.enum_class if input_type.enum_class is not None else input_type.enums
            return nw.Enum(categories)
        if isinstance(input_type, STRING_TYPES):
            return nw.String()
        if isinstance(input_type, sqltypes.Boolean):
            return nw.Boolean()
        if isinstance(input_type, sqltypes.SmallInteger):
            return nw.Int16()
        if isinstance(input_type, sqltypes.BigInteger):
            return nw.Int64()
        if isinstance(input_type, sqltypes.Integer):
            return nw.Int32()
        if isinstance(input_type, sqltypes.Double):
            return nw.Float64()
        if isinstance(input_type, (sqltypes.Float, sqltypes.REAL)):
            return nw.Float32()
        if isinstance(input_type, sqltypes.DECIMAL):
            return nw.Decimal()
        if isinstance(input_type, sqltypes.Numeric):
            # Safest option?
            return nw.Float64()
        if isinstance(input_type, sqltypes.Date):
            return nw.Date()
        if isinstance(input_type, sqltypes.DateTime):
            is_tz_aware, time_zone = input_type.timezone, get_anyschema_value_by_key(metadata, key="time_zone")
            if is_tz_aware and (time_zone is None):
                msg = (
                    "SQLAlchemy `DateTime(timezone=True)` does not specify a fixed timezone.\n\n"
                    "Hint: You can specify a timezone via `Column(..., info={'anyschema': {'time_zone': 'UTC'}})` "
                    "or `mapped_column(..., info={'anyschema': {'time_zone': 'UTC'}})`."
                )
                raise UnsupportedDTypeError(msg)
            if (not is_tz_aware) and (time_zone is not None):
                msg = f"SQLAlchemy `DateTime(timezone=False)` should not specify a fixed timezone, found {time_zone}"
                raise UnsupportedDTypeError(msg)
            return nw.Datetime(
                time_unit=get_anyschema_value_by_key(metadata, key="time_unit", default="us"), time_zone=time_zone
            )
        if isinstance(input_type, sqltypes.Time):
            return nw.Time()
        if isinstance(input_type, sqltypes.Interval):
            return nw.Duration()
        if isinstance(input_type, sqltypes._Binary):  # noqa: SLF001
            # All binary dtypes inherit from this one, namely: LargeBinary, BINARY, VARBINARY, BLOB, etc..
            return nw.Binary()

        # By exclusion, input_type is ARRAY.
        # ARRAY.item_type is a TypeEngine instance, which is also a valid FieldType
        # SQLAlchemy's type stubs don't provide full generic parameter information for item_type
        inner_type = self.pipeline.parse(input_type.item_type, constraints=constraints, metadata=metadata, strict=True)
        shape = input_type.dimensions
        return (
            nw.Array(inner=inner_type, shape=tuple(shape) if is_sequence_of(shape, int) else shape)
            if isinstance(shape, int) or is_sequence_of(shape, int)
            else nw.List(inner=inner_type)
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Mapping

import narwhals as nw
import pytest

from anyschema import AnySchema
from tests.conftest import (
    PydanticEventWithTimeMetadata,
    PydanticSpecialDatetimeWithMetadata,
    PydanticStudent,
)

if TYPE_CHECKING:
    from pydantic import BaseModel


@pytest.mark.parametrize(
    ("spec", "expected_schema"),
    [
        (
            PydanticStudent,
            {
                "name": nw.String(),
                "date_of_birth": nw.Date(),
                "age": nw.UInt64(),
                "classes": nw.List(nw.String()),
                "has_graduated": nw.Boolean(),
            },
        ),
        (
            PydanticEventWithTimeMetadata,
            {
                "name": nw.String(),
                "created_at": nw.Datetime("us"),
                "scheduled_at": nw.Datetime("us", time_zone="UTC"),
                "started_at": nw.Datetime("ms"),
                "completed_at": nw.Datetime("ns", time_zone="Europe/Berlin"),
            },
        ),
        (
            PydanticSpecialDatetimeWithMetadata,
            {
                "aware": nw.Datetime("us", time_zone="UTC"),
                "aware_ms": nw.Datetime("ms", time_zone="Asia/Tokyo"),
                "naive_ms": nw.Datetime("ms"),
                "past_utc": nw.Datetime("us", time_zone="UTC"),
                "future_ns": nw.Datetime("ns"),
            },
        ),
    ],
)
def test_pydantic_model(spec: type[BaseModel], expected_schema: Mapping[str, nw.dtypes.DType]) -> None:
    schema = AnySchema(spec=spec)
    nw_schema = schema._nw_schema
    assert nw_schema == nw.Schema(expected_schema)



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from anyschema import AnySchema

if TYPE_CHECKING:
    from anyschema.typing import Spec


@pytest.mark.parametrize(
    ("spec", "expected"),
    [
        ({}, ()),
        ({"only_field": str}, ("only_field",)),
        ({"id": int, "name": str, "age": int}, ("id", "name", "age")),
    ],
)
def test_names(spec: Spec, expected: tuple[str, ...]) -> None:
    schema = AnySchema(spec=spec)
    result = schema.names()

    assert result == expected



from __future__ import annotations

from typing import TYPE_CHECKING, Mapping

import narwhals as nw
import pytest

from anyschema import AnySchema
from tests.conftest import (
    AttrsAddressWithPydantic,
    AttrsDerived,
    AttrsEventWithTimeMetadata,
    AttrsPerson,
    AttrsPersonWithLiterals,
    create_missing_decorator_test_case,
)

if TYPE_CHECKING:
    from anyschema.typing import AttrsClassType


@pytest.mark.parametrize(
    ("spec", "expected_schema"),
    [
        (
            AttrsPerson,
            {
                "name": nw.String(),
                "age": nw.Int64(),
                "date_of_birth": nw.Date(),
                "is_active": nw.Boolean(),
                "classes": nw.List(nw.String()),
                "grades": nw.List(nw.Float64()),
            },
        ),
        (
            AttrsPersonWithLiterals,
            {
                "username": nw.String(),
                "role": nw.Enum(["admin", "user", "guest"]),
                "status": nw.Enum(["active", "inactive", "pending"]),
            },
        ),
        (
            AttrsAddressWithPydantic,
            {
                "street": nw.String(),
                "city": nw.String(),
                "zipcode": nw.Struct([nw.Field("zipcode", nw.UInt64())]),
            },
        ),
        (
            AttrsDerived,
            {
                "foo": nw.String(),
                "bar": nw.Int64(),
                "baz": nw.Float64(),
            },
        ),
        (
            AttrsEventWithTimeMetadata,
            {
                "name": nw.String(),
                "created_at": nw.Datetime("us"),
                "scheduled_at": nw.Datetime("us", time_zone="UTC"),
                "started_at": nw.Datetime("ms"),
                "completed_at": nw.Datetime("ns", time_zone="Europe/Berlin"),
            },
        ),
    ],
)
def test_attrs_class(spec: AttrsClassType, expected_schema: Mapping[str, nw.dtypes.DType]) -> None:
    schema = AnySchema(spec=spec)
    nw_schema = schema._nw_schema
    assert nw_schema == nw.Schema(expected_schema)


def test_attrs_class_missing_decorator_raises() -> None:
    child_cls, expected_msg = create_missing_decorator_test_case()
    with pytest.raises(AssertionError, match=expected_msg.replace("(", r"\(").replace(")", r"\)")):
        AnySchema(spec=child_cls)



"""Tests using actual pydantic-extra-types to verify derived type handling.

This module tests that PyTypeStep works with real types from the pydantic-extra-types library.
Note that some pydantic-extra-types require additional dependencies (like pycountry, phonenumbers).

References:
- https://docs.pydantic.dev/latest/api/pydantic_extra_types_country/
- https://docs.pydantic.dev/latest/api/pydantic_extra_types_phone_numbers/
- https://docs.pydantic.dev/latest/api/pydantic_extra_types_coordinate/
"""

from __future__ import annotations

from typing import Any

import narwhals as nw
import pytest
from pydantic_extra_types.coordinate import Latitude, Longitude
from pydantic_extra_types.country import (
    CountryAlpha2,
    CountryAlpha3,
    CountryNumericCode,
    CountryShortName,
)
from pydantic_extra_types.phone_numbers import PhoneNumber

from anyschema.parsers import ParserPipeline, PyTypeStep, UnionTypeStep


@pytest.fixture(scope="module")
def py_type_parser() -> PyTypeStep:
    """Create a PyTypeStep instance with pipeline set."""
    union_parser = UnionTypeStep()
    py_parser = PyTypeStep()
    _ = ParserPipeline([union_parser, py_parser])
    return py_parser


@pytest.mark.parametrize(
    ("input_type", "expected"),
    [
        # coordinate
        (Latitude, nw.Float64()),
        (Longitude, nw.Float64()),
        (list[Latitude], nw.List(nw.Float64())),
        (list[list[Latitude]], nw.List(nw.List(nw.Float64()))),
        (tuple[Longitude, Longitude], nw.Array(nw.Float64(), shape=2)),
        (tuple[Latitude, Latitude, Latitude], nw.Array(nw.Float64(), shape=3)),
        # country
        (CountryAlpha2, nw.String()),
        (CountryAlpha3, nw.String()),
        (CountryNumericCode, nw.String()),
        (CountryShortName, nw.String()),
        (list[CountryAlpha2], nw.List(nw.String())),
        (list[list[CountryAlpha2]], nw.List(nw.List(nw.String()))),
        # phone number
        (PhoneNumber, nw.String()),
        (list[PhoneNumber], nw.List(nw.String())),
        (tuple[PhoneNumber, PhoneNumber, PhoneNumber], nw.Array(nw.String(), shape=3)),
    ],
)
def test_pydantic_extra_types(py_type_parser: PyTypeStep, input_type: Any, expected: nw.dtypes.DType) -> None:
    result = py_type_parser.parse(input_type, (), {})
    assert result == expected



from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Optional

import hypothesis.strategies as st
import narwhals as nw
from annotated_types import Interval
from hypothesis import given
from pydantic import BaseModel, NegativeInt, NonNegativeInt, NonPositiveInt, PositiveInt

from tests.pydantic.utils import model_to_nw_schema

if TYPE_CHECKING:
    from anyschema.parsers import ParserPipeline


def test_parse_integer(auto_pipeline: ParserPipeline) -> None:
    class IntegerModel(BaseModel):
        # python integer type
        py_int: int
        py_int_optional: int | None
        py_int_or_none: int | None
        none_or_py_int: None | int

        # pydantic NonNegativeInt type
        non_negative: NonNegativeInt
        non_negative_optional: NonNegativeInt | None
        non_negative_or_none: NonNegativeInt | None
        none_or_non_negative: None | NonNegativeInt

        # pydantic NonPositiveInt type
        non_positive: NonPositiveInt
        non_positive_optional: NonPositiveInt | None
        non_positive_or_none: NonPositiveInt | None
        none_or_non_positive: None | NonPositiveInt

        # pydantic PositiveInt type
        positive: PositiveInt
        positive_optional: PositiveInt | None
        positive_or_none: PositiveInt | None
        none_or_positive: None | PositiveInt

        # pydantic NegativeInt type
        negative: NegativeInt
        negative_optional: NegativeInt | None
        negative_or_none: NegativeInt | None
        none_or_negative: None | NegativeInt

    schema = model_to_nw_schema(IntegerModel, pipeline=auto_pipeline)

    expected = {
        "py_int": nw.Int64(),
        "py_int_optional": nw.Int64(),
        "py_int_or_none": nw.Int64(),
        "none_or_py_int": nw.Int64(),
        "non_negative": nw.UInt64(),
        "non_negative_optional": nw.UInt64(),
        "non_negative_or_none": nw.UInt64(),
        "none_or_non_negative": nw.UInt64(),
        "non_positive": nw.Int64(),
        "non_positive_optional": nw.Int64(),
        "non_positive_or_none": nw.Int64(),
        "none_or_non_positive": nw.Int64(),
        "positive": nw.UInt64(),
        "positive_optional": nw.UInt64(),
        "positive_or_none": nw.UInt64(),
        "none_or_positive": nw.UInt64(),
        "negative": nw.Int64(),
        "negative_optional": nw.Int64(),
        "negative_or_none": nw.Int64(),
        "none_or_negative": nw.Int64(),
    }
    assert schema == expected


@given(lb=st.integers(-128, -2), ub=st.integers(2, 127))
def test_parse_to_int8(auto_pipeline: ParserPipeline, lb: int, ub: int) -> None:
    class Int8Model(BaseModel):
        x: Annotated[int, Interval(gt=lb, lt=ub)]
        y: Optional[Annotated[int, Interval(ge=lb, lt=ub)]]
        z: Annotated[int, Interval(gt=lb, le=ub)] | None
        w: None | Annotated[int, Interval(ge=lb, le=ub)]

    schema = model_to_nw_schema(Int8Model, pipeline=auto_pipeline)

    assert schema == {"x": nw.Int8(), "y": nw.Int8(), "z": nw.Int8(), "w": nw.Int8()}


@given(lb=st.integers(-32768, -129), ub=st.integers(129, 32767))
def test_parse_to_int16(auto_pipeline: ParserPipeline, lb: int, ub: int) -> None:
    class Int16Model(BaseModel):
        x: Annotated[int, Interval(gt=lb, lt=ub)]
        y: Optional[Annotated[int, Interval(ge=lb, lt=ub)]]
        z: Annotated[int, Interval(gt=lb, le=ub)] | None
        w: None | Annotated[int, Interval(ge=lb, le=ub)]

    schema = model_to_nw_schema(Int16Model, pipeline=auto_pipeline)

    assert schema == {"x": nw.Int16(), "y": nw.Int16(), "z": nw.Int16(), "w": nw.Int16()}


@given(lb=st.integers(-2147483648, -32769), ub=st.integers(32769, 2147483647))
def test_parse_to_int32(auto_pipeline: ParserPipeline, lb: int, ub: int) -> None:
    class Int32Model(BaseModel):
        x: Annotated[int, Interval(gt=lb, lt=ub)]
        y: Optional[Annotated[int, Interval(ge=lb, lt=ub)]]
        z: Annotated[int, Interval(gt=lb, le=ub)] | None
        w: None | Annotated[int, Interval(ge=lb, le=ub)]

    schema = model_to_nw_schema(Int32Model, pipeline=auto_pipeline)

    assert schema == {"x": nw.Int32(), "y": nw.Int32(), "z": nw.Int32(), "w": nw.Int32()}


@given(lb=st.integers(-9223372036854775808, -2147483649), ub=st.integers(2147483649, 9223372036854775808))
def test_parse_to_int64(auto_pipeline: ParserPipeline, lb: int, ub: int) -> None:
    class Int64Model(BaseModel):
        x: Annotated[int, Interval(gt=lb, lt=ub)]
        y: Optional[Annotated[int, Interval(ge=lb, lt=ub)]]
        z: Annotated[int, Interval(gt=lb, le=ub)] | None
        w: None | Annotated[int, Interval(ge=lb, le=ub)]

    schema = model_to_nw_schema(Int64Model, pipeline=auto_pipeline)

    assert schema == {"x": nw.Int64(), "y": nw.Int64(), "z": nw.Int64(), "w": nw.Int64()}


@given(ub=st.integers(1, 255))
def test_parse_to_uint8(auto_pipeline: ParserPipeline, ub: int) -> None:
    class UInt8Model(BaseModel):
        x: Annotated[int, Interval(gt=0, lt=ub)]
        y: Optional[Annotated[int, Interval(ge=0, lt=ub)]]
        z: Annotated[int, Interval(gt=0, le=ub)] | None
        w: None | Annotated[int, Interval(ge=0, le=ub)]

    schema = model_to_nw_schema(UInt8Model, pipeline=auto_pipeline)

    assert schema == {"x": nw.UInt8(), "y": nw.UInt8(), "z": nw.UInt8(), "w": nw.UInt8()}


@given(ub=st.integers(257, 65535))
def test_parse_to_uint16(auto_pipeline: ParserPipeline, ub: int) -> None:
    class UInt16Model(BaseModel):
        x: Annotated[int, Interval(gt=0, lt=ub)]
        y: Optional[Annotated[int, Interval(ge=0, lt=ub)]]
        z: Annotated[int, Interval(gt=0, le=ub)] | None
        w: None | Annotated[int, Interval(ge=0, le=ub)]

    schema = model_to_nw_schema(UInt16Model, pipeline=auto_pipeline)

    assert schema == {"x": nw.UInt16(), "y": nw.UInt16(), "z": nw.UInt16(), "w": nw.UInt16()}


@given(ub=st.integers(65537, 4294967295))
def test_parse_to_uint32(auto_pipeline: ParserPipeline, ub: int) -> None:
    class UInt32Model(BaseModel):
        x: Annotated[int, Interval(gt=0, lt=ub)]
        y: Optional[Annotated[int, Interval(ge=0, lt=ub)]]
        z: Annotated[int, Interval(gt=0, le=ub)] | None
        w: None | Annotated[int, Interval(ge=0, le=ub)]

    schema = model_to_nw_schema(UInt32Model, pipeline=auto_pipeline)

    assert schema == {"x": nw.UInt32(), "y": nw.UInt32(), "z": nw.UInt32(), "w": nw.UInt32()}


@given(ub=st.integers(4294967297, 18446744073709551615))
def test_parse_to_uint64(auto_pipeline: ParserPipeline, ub: int) -> None:
    class UInt64Model(BaseModel):
        x: Annotated[int, Interval(gt=0, lt=ub)]
        y: Optional[Annotated[int, Interval(ge=0, lt=ub)]]
        z: Annotated[int, Interval(gt=0, le=ub)] | None
        w: None | Annotated[int, Interval(ge=0, le=ub)]

    schema = model_to_nw_schema(UInt64Model, pipeline=auto_pipeline)

    assert schema == {"x": nw.UInt64(), "y": nw.UInt64(), "z": nw.UInt64(), "w": nw.UInt64()}


@given(value=st.integers(9223372036854775808))
def test_parse_to_int64_from_unbounded(auto_pipeline: ParserPipeline, value: int) -> None:
    class UnboundedModel(BaseModel):
        x: Annotated[int, Interval(lt=value)]

    schema = model_to_nw_schema(UnboundedModel, pipeline=auto_pipeline)

    assert schema == {"x": nw.Int64()}



from __future__ import annotations

import pytest

from anyschema import AnyField, AnySchema


def test_field_correct_attributes() -> None:
    schema = AnySchema(spec={"id": int, "name": str, "age": int | None})

    id_field = schema.field("id")
    assert isinstance(id_field, AnyField)
    assert id_field.name == "id"
    assert id_field.nullable is False

    age_field = schema.field("age")
    assert isinstance(age_field, AnyField)
    assert age_field.name == "age"
    assert age_field.nullable is True


def test_field_raises_keyerror_for_missing_field() -> None:
    schema = AnySchema(spec={"id": int, "name": str})

    with pytest.raises(KeyError):
        schema.field("non_existent")


def test_field_with_all_field_names() -> None:
    spec = {"id": int, "name": str, "age": int, "active": bool}
    schema = AnySchema(spec=spec)

    for field_name in spec:
        field = schema.field(field_name)
        assert field.name == field_name



from __future__ import annotations

from collections.abc import Iterable, Mapping, Sequence
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from enum import Enum
from types import NoneType
from typing import Any, Literal, TypedDict

import narwhals as nw
import pytest

from anyschema.exceptions import UnsupportedDTypeError
from anyschema.parsers import ParserPipeline, PyTypeStep, UnionTypeStep


class Color(Enum):
    """Example enum for testing."""

    RED = 1
    GREEN = 2
    BLUE = 3


class Status(Enum):
    """Another enum for testing."""

    ACTIVE = "active"
    INACTIVE = "inactive"


class CustomClass:
    pass


@pytest.fixture(scope="module")
def py_type_parser() -> PyTypeStep:
    """Create a PyTypeStep instance with pipeline set."""
    union_parser = UnionTypeStep()
    py_parser = PyTypeStep()
    _ = ParserPipeline([union_parser, py_parser])
    return py_parser


@pytest.mark.parametrize(
    ("input_type", "expected"),
    [
        (int, nw.Int64()),
        (float, nw.Float64()),
        (str, nw.String()),
        (bool, nw.Boolean()),
        (datetime, nw.Datetime("us")),
        (date, nw.Date()),
        (timedelta, nw.Duration()),
        (time, nw.Time()),
        (Decimal, nw.Decimal()),
        (bytes, nw.Binary()),
        (object, nw.Object()),
        (Color, nw.Enum(Color)),
        (Status, nw.Enum(Status)),
        (Literal["red", "green", "blue"], nw.Enum(("red", "green", "blue"))),
        (Literal[1, 2, 3], nw.Enum([1, 2, 3])),  # type: ignore[list-item]
        (Literal["active", "inactive"], nw.Enum(("active", "inactive"))),
        (Literal[True, False], nw.Enum([True, False])),  # type: ignore[list-item]
    ],
)
def test_parse_non_nested(py_type_parser: PyTypeStep, input_type: Any, expected: nw.dtypes.DType) -> None:
    result = py_type_parser.parse(input_type, (), {})
    assert result == expected


@pytest.mark.parametrize(
    ("input_type", "expected"),
    [
        (list[int], nw.List(nw.Int64())),
        (list[str], nw.List(nw.String())),
        (list[float], nw.List(nw.Float64())),
        (list[bool], nw.List(nw.Boolean())),
        (list[list[int]], nw.List(nw.List(nw.Int64()))),
        (list[tuple[int, int]], nw.List(nw.Array(nw.Int64(), shape=2))),
        (Sequence[int], nw.List(nw.Int64())),
        (Sequence[str | None], nw.List(nw.String())),
        (Iterable[int], nw.List(nw.Int64())),
        (Iterable[str], nw.List(nw.String())),
        (Sequence[Sequence[str]], nw.List(nw.List(nw.String()))),
        (tuple[int | None, ...], nw.List(nw.Int64())),
        (tuple[int, int, int], nw.Array(nw.Int64(), shape=3)),
        (tuple[str, str], nw.Array(nw.String(), shape=2)),
        (list, nw.List(nw.Object())),
        (tuple, nw.List(nw.Object())),
        (Sequence, nw.List(nw.Object())),
        (Iterable, nw.List(nw.Object())),
    ],
)
def test_parse_nested(py_type_parser: PyTypeStep, input_type: Any, expected: nw.dtypes.DType) -> None:
    result = py_type_parser.parse(input_type, (), {})
    assert result == expected


@pytest.mark.parametrize("input_type", [tuple[int, str], tuple[int, str, float]])
def test_parse_heterogeneous_tuple_raises(py_type_parser: PyTypeStep, input_type: Any) -> None:
    with pytest.raises(UnsupportedDTypeError, match="Tuple with mixed types is not supported"):
        py_type_parser.parse(input_type, (), {})


@pytest.mark.parametrize(("input_type", "expected"), [(int, nw.Int64()), (list[int], nw.List(nw.Int64()))])
def test_parse_with_metadata(py_type_parser: PyTypeStep, input_type: Any, expected: nw.dtypes.DType) -> None:
    result = py_type_parser.parse(input_type, constraints=("some", "constraints"), metadata={"key": "value"})
    assert result == expected


@pytest.mark.parametrize(
    ("metadata", "expected"),
    [
        ({}, nw.Datetime("us")),
        ({"anyschema": {"time_unit": "ms"}}, nw.Datetime("ms")),
        ({"anyschema": {"time_unit": "ns"}}, nw.Datetime("ns")),
        ({"anyschema": {"time_zone": "UTC"}}, nw.Datetime("us", time_zone="UTC")),
        ({"anyschema": {"time_zone": "Europe/Rome"}}, nw.Datetime("us", time_zone="Europe/Rome")),
        ({"anyschema": {"time_unit": "ms", "time_zone": "UTC"}}, nw.Datetime("ms", time_zone="UTC")),
        (
            {"anyschema": {"time_unit": "ns", "time_zone": "America/New_York"}},
            nw.Datetime("ns", time_zone="America/New_York"),
        ),
    ],
)
def test_parse_datetime_with_time_metadata(
    py_type_parser: PyTypeStep, metadata: dict[str, Any], expected: nw.dtypes.DType
) -> None:
    """Test that datetime parses correctly with anyschema/time_zone and anyschema/time_unit metadata."""
    result = py_type_parser.parse(datetime, constraints=(), metadata=metadata)
    assert result == expected


@pytest.mark.parametrize("input_type", [CustomClass, NoneType, set[int], frozenset])
def test_unsupported_type(py_type_parser: PyTypeStep, input_type: Any) -> None:
    result = py_type_parser.parse(input_type, (), {})
    assert result is None


class PersonTypedDict(TypedDict):
    """Example TypedDict for testing."""

    name: str
    age: int


class AddressTypedDict(TypedDict):
    """Nested TypedDict for testing."""

    street: str
    city: str
    zipcode: int


class PersonWithAddressTypedDict(TypedDict):
    """TypedDict with nested TypedDict for testing."""

    name: str
    address: AddressTypedDict


@pytest.mark.parametrize(
    ("input_type", "expected"),
    [
        # Plain dict -> empty Struct
        (dict, nw.Struct([])),
        # dict[K, V] -> empty Struct (we don't use the value type for now)
        (dict[str, int], nw.Struct([])),
        (dict[str, str], nw.Struct([])),
        (Mapping[str, int], nw.Struct([])),
        # TypedDict -> Struct with fields
        (
            PersonTypedDict,
            nw.Struct([nw.Field(name="name", dtype=nw.String()), nw.Field(name="age", dtype=nw.Int64())]),
        ),
        (
            AddressTypedDict,
            nw.Struct(
                [
                    nw.Field(name="street", dtype=nw.String()),
                    nw.Field(name="city", dtype=nw.String()),
                    nw.Field(name="zipcode", dtype=nw.Int64()),
                ]
            ),
        ),
        # Nested TypedDict
        (
            PersonWithAddressTypedDict,
            nw.Struct(
                [
                    nw.Field(name="name", dtype=nw.String()),
                    nw.Field(
                        name="address",
                        dtype=nw.Struct(
                            [
                                nw.Field(name="street", dtype=nw.String()),
                                nw.Field(name="city", dtype=nw.String()),
                                nw.Field(name="zipcode", dtype=nw.Int64()),
                            ]
                        ),
                    ),
                ]
            ),
        ),
    ],
)
def test_parse_dict_types(py_type_parser: PyTypeStep, input_type: Any, expected: nw.dtypes.DType) -> None:
    result = py_type_parser.parse(input_type, (), {})
    assert result == expected



from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Optional

import narwhals as nw
from annotated_types import Interval, Len
from pydantic import BaseModel

from tests.pydantic.utils import model_to_nw_schema

if TYPE_CHECKING:
    from anyschema.parsers import ParserPipeline


def test_parse_list_optional_outer(auto_pipeline: ParserPipeline) -> None:
    class ListModel(BaseModel):
        # python list[...] type
        py_list: list[int]
        py_list_optional: list[str] | None
        py_list_or_none: list[float] | None
        none_or_py_list: None | list[bool]

        # pydantic conlist type
        con_list: Annotated[list[int], Len(min_length=2)]
        con_list_optional: Optional[Annotated[list[str], Len(max_length=6)]]
        con_list_or_none: Annotated[list[float], Len(0)] | None
        none_or_con_list: None | Annotated[list[bool], Len(0)]

    schema = model_to_nw_schema(ListModel, pipeline=auto_pipeline)
    expected = {
        "py_list": nw.List(nw.Int64()),
        "py_list_optional": nw.List(nw.String()),
        "py_list_or_none": nw.List(nw.Float64()),
        "none_or_py_list": nw.List(nw.Boolean()),
        "con_list": nw.List(nw.Int64()),
        "con_list_optional": nw.List(nw.String()),
        "con_list_or_none": nw.List(nw.Float64()),
        "none_or_con_list": nw.List(nw.Boolean()),
    }
    assert schema == expected


def test_parse_list_optional_inner(auto_pipeline: ParserPipeline) -> None:
    class ListModel(BaseModel):
        # python list[...] type
        py_list_optional: list[str | None]
        py_list_or_none: list[float | None] | None
        none_or_py_list: list[None | bool]

        # pydantic conlist type
        con_list_optional: Annotated[list[Optional[int]], Len(min_length=2)]
        con_list_or_none: Annotated[list[str | None], Len(max_length=6)]
        none_or_con_list: Annotated[list[None | float], Len(0)]

    schema = model_to_nw_schema(ListModel, pipeline=auto_pipeline)
    expected = {
        "py_list_optional": nw.List(nw.String()),
        "py_list_or_none": nw.List(nw.Float64()),
        "none_or_py_list": nw.List(nw.Boolean()),
        "con_list_optional": nw.List(nw.Int64()),
        "con_list_or_none": nw.List(nw.String()),
        "none_or_con_list": nw.List(nw.Float64()),
    }
    assert schema == expected


def test_parse_list_optional_outer_and_inner(auto_pipeline: ParserPipeline) -> None:
    class ListModel(BaseModel):
        # python list[...] type
        py_list_optional_optional: list[int | None] | None
        py_list_optional_none: list[str | None] | None
        py_list_none_optional: list[float | None] | None
        py_list_none_none: list[None | bool] | None

        # pydantic conlist type
        con_list_optional_optional: Optional[Annotated[list[Optional[int]], Len(min_length=2)]]
        con_list_optional_none: Annotated[list[Optional[str]], Len(max_length=6)] | None
        con_list_none_optional: Optional[Annotated[list[float | None], Len(0)]]
        con_list_none_none: Annotated[list[None | bool], Len(0)] | None

    schema = model_to_nw_schema(ListModel, pipeline=auto_pipeline)
    expected = {
        "py_list_optional_optional": nw.List(nw.Int64()),
        "py_list_optional_none": nw.List(nw.String()),
        "py_list_none_optional": nw.List(nw.Float64()),
        "py_list_none_none": nw.List(nw.Boolean()),
        "con_list_optional_optional": nw.List(nw.Int64()),
        "con_list_optional_none": nw.List(nw.String()),
        "con_list_none_optional": nw.List(nw.Float64()),
        "con_list_none_none": nw.List(nw.Boolean()),
    }
    assert schema == expected


def test_parse_conlist_conint(auto_pipeline: ParserPipeline) -> None:
    class ListModel(BaseModel):
        # python list[...] type
        py_list_int8: list[Annotated[int, Interval(gt=-64, lt=64)]] | None
        py_list_uint8: list[Annotated[int, Interval(gt=0, lt=64)] | None]

        # pydantic conlist type
        con_list_int8: Annotated[list[None | Annotated[int, Interval(gt=-64, lt=64)]], Len(0)]
        con_list_uint8: Annotated[list[Optional[Annotated[int, Interval(gt=0, lt=64)]]], Len(0)]

    schema = model_to_nw_schema(ListModel, pipeline=auto_pipeline)
    expected = {
        "py_list_int8": nw.List(nw.Int8()),
        "py_list_uint8": nw.List(nw.UInt8()),
        "con_list_int8": nw.List(nw.Int8()),
        "con_list_uint8": nw.List(nw.UInt8()),
    }
    assert schema == expected



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from sqlalchemy import BigInteger, Float, Integer, String
from sqlalchemy.types import DateTime, TypeEngine

from anyschema.adapters import sqlalchemy_adapter
from tests.conftest import (
    EventORMWithTimeMetadata,
    SimpleUserORM,
    UserWithTypesORM,
    event_table_with_time_metadata,
    event_table_with_tz_aware,
    numeric_table,
    user_table,
)

if TYPE_CHECKING:
    from collections.abc import Iterable

    from anyschema.typing import FieldSpec, SQLAlchemyTableType


def assert_result_equal(result: Iterable[FieldSpec], expected: Iterable[FieldSpec]) -> None:
    """Helper function that takes care of the fact that sqlalchemy types comparison results in False.

    Namely Integer()==Integer() -> False, therefore we compare their string representation instead.
    """
    for left, right in zip(result, expected, strict=True):
        for lval, rval in zip(left, right, strict=True):
            if isinstance(lval, TypeEngine) and isinstance(rval, TypeEngine):
                assert str(lval) == str(rval), f"{left} != {right}"
            else:
                assert lval == rval, f"{left} != {right}"


@pytest.mark.parametrize(
    ("spec", "expected"),
    [
        (
            SimpleUserORM,
            [
                ("id", Integer(), (), {"anyschema": {"nullable": False}}),
                ("name", String(), (), {"anyschema": {"nullable": True}}),
            ],
        ),
        (
            UserWithTypesORM,
            [
                ("id", Integer(), (), {"anyschema": {"nullable": False}}),
                ("name", String(50), (), {"anyschema": {"nullable": False}}),
                ("age", Integer(), (), {"anyschema": {"nullable": True}}),
                ("score", Float(), (), {"anyschema": {"nullable": True}}),
            ],
        ),
        (
            user_table,
            [
                (
                    "id",
                    Integer(),
                    (),
                    {"anyschema": {"nullable": False, "description": "Primary key"}},
                ),
                ("name", String(50), (), {"anyschema": {"nullable": True}}),
                (
                    "age",
                    Integer(),
                    (),
                    {"anyschema": {"nullable": True, "description": "User age"}},
                ),
                ("email", String(100), (), {"anyschema": {"nullable": True}}),
            ],
        ),
        (
            numeric_table,
            [
                ("int_col", Integer(), (), {"anyschema": {"nullable": True}}),
                ("bigint_col", BigInteger(), (), {"anyschema": {"nullable": True}}),
                ("string_col", String(100), (), {"anyschema": {"nullable": True}}),
                ("float_col", Float(), (), {"anyschema": {"nullable": True}}),
            ],
        ),
    ],
)
def test_sqlalchemy_adapter(spec: SQLAlchemyTableType, expected: Iterable[FieldSpec]) -> None:
    result = sqlalchemy_adapter(spec)
    assert_result_equal(result, expected)


def test_sqlalchemy_adapter_with_time_metadata_table() -> None:
    """Test that SQLAlchemy Table adapter correctly extracts time metadata from column.info."""
    result = tuple(sqlalchemy_adapter(event_table_with_time_metadata))
    expected: Iterable[FieldSpec] = (
        ("id", Integer(), (), {"anyschema": {"nullable": False}}),
        ("name", String(100), (), {"anyschema": {"nullable": True}}),
        ("created_at", DateTime(), (), {"anyschema": {"nullable": True}}),
        ("scheduled_at", DateTime(timezone=True), (), {"anyschema": {"nullable": True, "time_zone": "UTC"}}),
        ("started_at", DateTime(), (), {"anyschema": {"nullable": True, "time_unit": "ms"}}),
        (
            "completed_at",
            DateTime(timezone=True),
            (),
            {"anyschema": {"nullable": True, "time_zone": "Europe/Berlin", "time_unit": "ns"}},
        ),
    )
    assert_result_equal(result, expected)


def test_sqlalchemy_adapter_with_time_metadata_orm() -> None:
    """Test that SQLAlchemy ORM adapter correctly extracts time metadata from mapped_column info."""
    result = tuple(sqlalchemy_adapter(EventORMWithTimeMetadata))
    expected: Iterable[FieldSpec] = (
        ("id", Integer(), (), {"anyschema": {"nullable": False}}),
        ("name", String(), (), {"anyschema": {"nullable": False}}),
        ("created_at", DateTime(), (), {"anyschema": {"nullable": False}}),
        ("scheduled_at", DateTime(timezone=True), (), {"anyschema": {"nullable": False, "time_zone": "UTC"}}),
        ("started_at", DateTime(), (), {"anyschema": {"nullable": False, "time_unit": "ms"}}),
        (
            "completed_at",
            DateTime(timezone=True),
            (),
            {"anyschema": {"nullable": False, "time_zone": "Europe/Berlin", "time_unit": "ns"}},
        ),
    )
    assert_result_equal(result, expected)


def test_sqlalchemy_adapter_with_tz_aware_datetime() -> None:
    """Test that SQLAlchemy adapter correctly extracts timezone-aware datetime metadata."""
    result = tuple(sqlalchemy_adapter(event_table_with_tz_aware))
    expected: Iterable[FieldSpec] = (
        ("id", Integer(), (), {"anyschema": {"nullable": False}}),
        ("timestamp_utc", DateTime(timezone=True), (), {"anyschema": {"nullable": True, "time_zone": "UTC"}}),
        (
            "timestamp_berlin",
            DateTime(timezone=True),
            (),
            {"anyschema": {"nullable": True, "time_zone": "Europe/Berlin", "time_unit": "ms"}},
        ),
    )
    assert_result_equal(result, expected)



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Mapping, TypedDict

import narwhals as nw
import pytest
from pydantic import BaseModel, PositiveInt

from anyschema import AnySchema

if TYPE_CHECKING:
    from anyschema.typing import TypedDictType


class PersonTypedDict(TypedDict):
    """Simple TypedDict for testing."""

    name: str
    age: int
    is_active: bool


class AddressTypedDict(TypedDict):
    """Nested TypedDict for testing."""

    street: str
    city: str
    zipcode: int


class PersonWithAddressTypedDict(TypedDict):
    """TypedDict with nested TypedDict for testing."""

    name: str
    age: int
    address: AddressTypedDict


class StudentTypedDict(TypedDict):
    """TypedDict with list field for testing."""

    name: str
    age: int
    classes: list[str]
    grades: list[float]


class UserTypedDict(TypedDict):
    """TypedDict with Literal fields for testing."""

    username: str
    role: Literal["admin", "user", "guest"]
    status: Literal["active", "inactive", "pending"]
    age: int


class ConfigTypedDict(TypedDict):
    """TypedDict with mixed Literal types for testing."""

    name: str
    log_level: Literal["debug", "info", "warning", "error"]
    max_retries: Literal[1, 2, 3, 5, 10]
    enabled: Literal[True, False]


class ZipcodeModel(BaseModel):
    zipcode: PositiveInt


class AddressTypedDictWithZipcodeModel(TypedDict):
    """TypedDict with Nested pydantic model for testing."""

    street: str
    city: str
    zipcode: ZipcodeModel


@pytest.mark.parametrize(
    ("spec", "expected_schema"),
    [
        (PersonTypedDict, {"name": nw.String(), "age": nw.Int64(), "is_active": nw.Boolean()}),
        (
            PersonWithAddressTypedDict,
            {
                "name": nw.String(),
                "age": nw.Int64(),
                "address": nw.Struct(
                    [
                        nw.Field("street", nw.String()),
                        nw.Field("city", nw.String()),
                        nw.Field("zipcode", nw.Int64()),
                    ]
                ),
            },
        ),
        (
            StudentTypedDict,
            {"name": nw.String(), "age": nw.Int64(), "classes": nw.List(nw.String()), "grades": nw.List(nw.Float64())},
        ),
        (
            UserTypedDict,
            {
                "username": nw.String(),
                "role": nw.Enum(["admin", "user", "guest"]),
                "status": nw.Enum(["active", "inactive", "pending"]),
                "age": nw.Int64(),
            },
        ),
        (
            ConfigTypedDict,
            {
                "name": nw.String(),
                "log_level": nw.Enum(["debug", "info", "warning", "error"]),
                "max_retries": nw.Enum([1, 2, 3, 5, 10]),  # type: ignore[list-item]
                "enabled": nw.Enum([True, False]),  # type: ignore[list-item]
            },
        ),
        (
            AddressTypedDictWithZipcodeModel,
            {
                "street": nw.String(),
                "city": nw.String(),
                "zipcode": nw.Struct([nw.Field("zipcode", nw.UInt64())]),
            },
        ),
    ],
)
def test_typed_dict(spec: TypedDictType, expected_schema: Mapping[str, nw.dtypes.DType]) -> None:
    schema = AnySchema(spec=spec)
    nw_schema = schema._nw_schema
    assert nw_schema == nw.Schema(expected_schema)



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from anyschema import AnySchema
from tests.conftest import DataclassEventWithTimeMetadata, PydanticStudent, user_table

if TYPE_CHECKING:
    from anyschema.typing import Spec


@pytest.mark.parametrize(
    ("spec", "expected"),
    [
        (PydanticStudent, ("Student full name", None, "Student age in years", None, None)),
        (DataclassEventWithTimeMetadata, ("Event name", None, "Scheduled time", None, None)),
        (user_table, ("Primary key", None, "User age", None)),
    ],
)
def test_descriptions_named_false(spec: Spec, expected: tuple[str | None, ...]) -> None:
    schema = AnySchema(spec=spec)
    result = schema.descriptions(named=False)

    assert result == expected


@pytest.mark.parametrize(
    ("spec", "expected"),
    [
        (
            PydanticStudent,
            {
                "name": "Student full name",
                "date_of_birth": None,
                "age": "Student age in years",
                "classes": None,
                "has_graduated": None,
            },
        ),
        (
            DataclassEventWithTimeMetadata,
            {
                "name": "Event name",
                "created_at": None,
                "scheduled_at": "Scheduled time",
                "started_at": None,
                "completed_at": None,
            },
        ),
        (user_table, {"id": "Primary key", "name": None, "age": "User age", "email": None}),
    ],
)
def test_descriptions_named_true(spec: Spec, expected: dict[str, str | None]) -> None:
    schema = AnySchema(spec=spec)
    result = schema.descriptions(named=True)

    assert result == expected



from __future__ import annotations

from typing import TYPE_CHECKING, Mapping

import narwhals as nw
import pytest

from anyschema import AnySchema
from tests.conftest import DataclassEventWithTimeMetadata

if TYPE_CHECKING:
    from anyschema.typing import DataclassType


@pytest.mark.parametrize(
    ("spec", "expected_schema"),
    [
        (
            DataclassEventWithTimeMetadata,
            {
                "name": nw.String(),
                "created_at": nw.Datetime("us"),
                "scheduled_at": nw.Datetime("us", time_zone="UTC"),
                "started_at": nw.Datetime("ms"),
                "completed_at": nw.Datetime("ns", time_zone="Europe/Berlin"),
            },
        ),
    ],
)
def test_dataclass(spec: DataclassType, expected_schema: Mapping[str, nw.dtypes.DType]) -> None:
    schema = AnySchema(spec=spec)
    nw_schema = schema._nw_schema
    assert nw_schema == nw.Schema(expected_schema)
